from milcapy.core.node import Node
from milcapy.section.section import Section
import numpy as np
from milcapy.loads.load import DistributedLoad
from typing import Dict, Optional
from milcapy.utils.geometry import angle_x_axis

class TrussElement:
    """
    Elemento de armadura 3dof por nodo.
    """
    def __init__(
        self,
        id: int,
        node_i: Node,
        node_j: Node,
        section: Section,
    ) -> None:
        """
        Inicializa el elemento de armadura 3dof por nodo.

        Args:
            id (int): Identificador del elemento.
            node_i (Node): Primer nodo.
            node_j (Node): Segundo nodo.
            section (Section): Sección del elemento tipo area (shell).
        """
        self.id = id
        self.node_i = node_i
        self.node_j = node_j
        self.section = section
        self.E = section.E()
        self.A = section.A()
        self.dofs = np.concatenate((node_i.dofs[:2], node_j.dofs[:2]))

        self.distributed_load: Dict[str, DistributedLoad] = {}  # {pattern_name: DistributedLoad}
        self.displacements = None

        self.current_load_pattern: Optional[str] = None
        self.la = None; self.lb = None

    def length(self) -> float:
        """
        Longitud del miembro.
        """
        return (self.node_i.vertex.distance_to(self.node_j.vertex))

    def angle_x(self) -> float:
        """
        Ángulo del armadura respecto al eje X del sistema global.
        """
        return angle_x_axis(
            self.node_j.vertex.x - self.node_i.vertex.x,
            self.node_j.vertex.y - self.node_i.vertex.y
        )


    def transformation_matrix(self) -> np.ndarray:
        """
        Matriz de transformación del sistema local a global.
        """
        L = self.length()
        cx = (self.node_j.vertex.x - self.node_i.vertex.x) / L
        cy = (self.node_j.vertex.y - self.node_i.vertex.y) / L
        T = np.array([
            [cx, cy, 0, 0],
            [0, 0, cx, cy],
        ])
        return T

    def k_local(self) -> np.ndarray:
        """
        Matriz de rigidez local del elemento.
        """
        return self.E*self.A/self.length()*np.array([[1, -1], [-1, 1]])

    def global_stiffness_matrix(self) -> np.ndarray:
        """
        Matriz de rigidez global del elemento.
        """
        return self.transformation_matrix().T @ self.k_local() @ self.transformation_matrix()

    def global_load_vector(self) -> np.ndarray:
        """
        Vector de cargas global del elemento.
        """
        return self.transformation_matrix().T @ self.q_local()

    def set_current_load_pattern(self, load_pattern_name: str) -> None:
        """
        Establece el patrón de carga actual del miembro.
        """
        self.current_load_pattern = load_pattern_name

    def set_distributed_load(self, load: DistributedLoad) -> None:
        """
        Asigna una carga distribuida al miembro para el patrón de carga actual.
        """
        if self.current_load_pattern is None:
            raise ValueError("Debe establecer un patrón de carga actual antes de asignar una carga distribuida.")
        self.distributed_load[self.current_load_pattern] = load

    def get_distributed_load(self, load_pattern_name: str) -> Optional["DistributedLoad"]:
        """
        Obtiene la carga distribuida para el patrón de carga actual.
        """
        load = self.distributed_load.get(load_pattern_name, None)
        if load is None:
            return DistributedLoad()
        return load

    def local_stiffness_matrix(self) -> np.ndarray:
        """
        Matriz de rigidez local del elemento.
        """
        return self.k_local()

    def q_local(self) -> np.ndarray:
        """
        Vector de cargas distribuidas equivalentes en sistema local.
        """
        load = self.get_distributed_load(self.current_load_pattern)
        vector = np.array([
            2*load.p_i + load.p_j,
            2*load.p_j + load.p_i,
        ])
        return vector

    def global_load_vector(self) -> np.ndarray:
        """
        Vector de cargas global del elemento.
        """
        return self.transformation_matrix().T @ self.q_local()


from math import nan
from typing import Optional, Dict, Union, Tuple, List
import numpy as np
from abc import ABC, abstractmethod

class Load(ABC):
    """
    Clase base abstracta que define la interfaz común para diferentes tipos de cargas estructurales.

    Esta clase implementa operaciones básicas como comparación, inversión y conversión,
    que son comunes a todos los tipos de cargas.

    Attributes:
        id (Optional[int]): Identificador único de la carga. Por defecto es None.
    """

    def __init__(
        self,
        id: Optional[int] = None
        ) -> None:

        """
        Inicializa una nueva instancia de carga.

        Args:
            id: Identificador único opcional para la carga.
        """

        self.id = id

    @property
    @abstractmethod
    def components(
        self
        ) -> np.ndarray:
        """
        Obtiene los componentes de la carga como un array NumPy.

        Returns:
            np.ndarray: Array con los componentes de la carga.
        """
        pass

    @abstractmethod
    def to_dict(
        self
        ) -> Dict[str, Union[int, float, None]]:
        """
        Convierte la carga a un diccionario serializable.

        Returns:
            Dict[str, Union[int, float, None]]: Diccionario con los atributos de la carga.
        """
        pass

    def __eq__(
        self,
        other: object
        ) -> bool:
        """
        Compara si dos cargas son aproximadamente iguales.

        Args:
            other: Objeto a comparar.

        Returns:
            bool: True si las cargas son aproximadamente iguales, False en caso contrario.
        """
        if not isinstance(other, Load):
            return NotImplemented
        return np.allclose(self.components, other.components)

    def __neg__(
        self
        ) -> 'Load':
        """
        Devuelve una nueva instancia con todos los componentes de carga invertidos.

        Returns:
            Load: Nueva instancia con cargas negativas.
        """
        return self.__class__(*(-self.components), id=self.id)

    def __pos__(
        self
        ) -> 'Load':
        """
        Devuelve la misma instancia sin modificaciones (operador unario +).

        Returns:
            Load: La misma instancia sin cambios.
        """
        return self


class PointLoad(Load):
    """
    Representa una carga puntual en un sistema estructural 2D.

    Esta clase modela una carga concentrada con:
    - Fuerza de corte perpendicular al eje de la viga (fx)
    - Fuerza axial a lo largo del eje x de la viga (fy)
    - Momento que genera rotación alrededor del eje z (mz)

    Attributes:
        fx (float): Fuerza de corte perpendicular al eje de la viga.
        fy (float): Fuerza axial a lo largo del eje x de la viga.
        mz (float): Momento alrededor del eje z.
        id (Optional[int]): Identificador único de la carga.
    """

    __slots__ = ('fx', 'fy', 'mz', 'id')
    ZERO = np.zeros(3, dtype=np.float64)

    def __init__(
        self,
        fx: float = 0.0,
        fy: float = 0.0,
        mz: float = 0.0,
        id: Optional[int] = None
        ) -> None:
        """
        Inicializa una nueva carga puntual.

        Args:
            fx: Fuerza de corte perpendicular al eje de la viga.
            fy: Fuerza axial a lo largo del eje x de la viga.
            mz: Momento que genera rotación alrededor del eje z.
            id: Identificador único opcional.
        """
        super().__init__(id)
        self.fx = float(fx)
        self.fy = float(fy)
        self.mz = float(mz)

    @property
    def components(
        self
        ) -> np.ndarray:
        """
        Obtiene los componentes de la carga como un array NumPy.

        Returns:
            np.ndarray: Array con [fx, fy, mz].
        """
        return np.array([self.fx, self.fy, self.mz], dtype=np.float64)

    def to_dict(
        self
        ) -> Dict[str, Union[int, float, None]]:
        """
        Convierte la carga puntual a un diccionario serializable.

        Returns:
            Dict[str, Union[int, float, None]]: Diccionario con los componentes de la carga.
        """
        return {"fx": self.fx, "fy": self.fy, "mz": self.mz, "id": self.id}

    def __add__(
        self,
        other: "PointLoad"
        ) -> "PointLoad":
        """
        Suma dos cargas puntuales.

        Args:
            other: Otra carga puntual a sumar.

        Returns:
            PointLoad: Nueva carga puntual resultante de la suma.

        Raises:
            TypeError: Si other no es una instancia de PointLoad.
        """
        if not isinstance(other, PointLoad):
            return NotImplemented
        return PointLoad(
            self.fx + other.fx,
            self.fy + other.fy,
            self.mz + other.mz,
            self.id or other.id
        )

    def __sub__(
        self,
        other: "PointLoad"
        ) -> "PointLoad":
        """
        Resta dos cargas puntuales.

        Args:
            other: Carga puntual a restar.

        Returns:
            PointLoad: Nueva carga puntual resultante de la resta.

        Raises:
            TypeError: Si other no es una instancia de PointLoad.
        """
        if not isinstance(other, PointLoad):
            return NotImplemented
        return PointLoad(
            self.fx - other.fx,
            self.fy - other.fy,
            self.mz - other.mz,
            self.id or other.id
        )

    def __mul__(
        self,
        scalar: Union[float, int]
        ) -> "PointLoad":
        """
        Multiplica la carga por un escalar.

        Args:
            scalar: Factor de escala.

        Returns:
            PointLoad: Nueva carga puntual escalada.

        Raises:
            TypeError: Si scalar no es un número.
        """
        if not isinstance(scalar, (float, int)):
            return NotImplemented
        return PointLoad(
            self.fx * scalar,
            self.fy * scalar,
            self.mz * scalar,
            self.id
        )

    __rmul__ = __mul__

    def __truediv__(
        self,
        scalar: Union[float, int]
        ) -> "PointLoad":
        """
        Divide la carga por un escalar.

        Args:
            scalar: Divisor.

        Returns:
            PointLoad: Nueva carga puntual dividida.

        Raises:
            TypeError: Si scalar no es un número.
            ZeroDivisionError: Si scalar es cero.
        """
        if not isinstance(scalar, (float, int)):
            return NotImplemented
        if scalar == 0:
            raise ZeroDivisionError("No se puede dividir por cero.")
        return PointLoad(
            self.fx / scalar,
            self.fy / scalar,
            self.mz / scalar,
            self.id
        )


class DistributedLoad(Load):
    """
    Representa una carga distribuida en un elemento estructural 2D.

    Esta clase modela cargas distribuidas con valores iniciales (i) y finales (j):
    - q: Carga distribuida perpendicular al eje de la viga
    - p: Carga distribuida axial a lo largo del eje de la viga
    - m: Momento distribuido que genera rotación alrededor del eje z

    Attributes:
        q_i (float): Carga de corte inicial, perpendicular al eje de la viga, eje y.
        q_j (float): Carga de corte final, perpendicular al eje de la viga, eje y.
        p_i (float): Carga axial inicial, a lo largo del eje x de la viga.
        p_j (float): Carga axial final, a lo largo del eje x de la viga.
        m_i (float): Momento inicial alrededor del eje z.
        m_j (float): Momento final alrededor del eje z.
        id (Optional[int]): Identificador único de la carga.
    """

    __slots__ = ('q_i', 'q_j', 'p_i', 'p_j', 'm_i', 'm_j', 'id')
    ZERO = np.zeros(6, dtype=np.float64)

    def __init__(
        self,
        q_i: float = 0.0,
        q_j: float = 0.0,
        p_i: float = 0.0,
        p_j: float = 0.0,
        m_i: float = 0.0,
        m_j: float = 0.0,
        id: Optional[int] = None
        ) -> None:
        """
        Inicializa una nueva carga distribuida.

        Args:
            q_i: Carga de corte inicial, perpendicular al eje de la viga.
            q_j: Carga de corte final, perpendicular al eje de la viga.
            p_i: Carga axial inicial, a lo largo del eje x de la viga.
            p_j: Carga axial final, a lo largo del eje x de la viga.
            m_i: Momento inicial alrededor del eje z.
            m_j: Momento final alrededor del eje z.
            id: Identificador único opcional.

        Raises:
            TypeError: Si alguno de los valores no es un número real.
        """
        super().__init__(id)

        # Validación de tipos y conversión
        load_params = {'q_i': q_i, 'q_j': q_j, 'p_i': p_i,
                    'p_j': p_j, 'm_i': m_i, 'm_j': m_j}

        for name, value in load_params.items():
            if not isinstance(value, (int, float)):
                raise TypeError(f"El parámetro {name} debe ser un número real")
            setattr(self, name, float(value))

    @property
    def components(
        self
        ) -> np.ndarray:
        """
        Obtiene los componentes de la carga como un array NumPy.

        Returns:
            np.ndarray: Array con [q_i, q_j, p_i, p_j, m_i, m_j].
        """
        return np.array([self.q_i, self.q_j, self.p_i, self.p_j, self.m_i, self.m_j],
                       dtype=np.float64)

    def to_dict(
        self
        ) -> Dict[str, Union[int, float, None]]:
        """
        Convierte la carga distribuida a un diccionario serializable.

        Returns:
            Dict[str, Union[int, float, None]]: Diccionario con los componentes de la carga.
        """
        return {
            "q_i": self.q_i, "q_j": self.q_j,
            "p_i": self.p_i, "p_j": self.p_j,
            "m_i": self.m_i, "m_j": self.m_j,
            "id": self.id
        }

    def __add__(
        self,
        other: "DistributedLoad"
        ) -> "DistributedLoad":
        """
        Suma dos cargas distribuidas.

        Args:
            other: Otra carga distribuida a sumar.

        Returns:
            DistributedLoad: Nueva carga distribuida resultante de la suma.

        Raises:
            TypeError: Si other no es una instancia de DistributedLoad.
        """
        if not isinstance(other, DistributedLoad):
            return NotImplemented
        return DistributedLoad(
            self.q_i + other.q_i,
            self.q_j + other.q_j,
            self.p_i + other.p_i,
            self.p_j + other.p_j,
            self.m_i + other.m_i,
            self.m_j + other.m_j,
            self.id or other.id
        )

    def __sub__(
        self,
        other: "DistributedLoad"
        ) -> "DistributedLoad":
        """
        Resta dos cargas distribuidas.

        Args:
            other: Carga distribuida a restar.

        Returns:
            DistributedLoad: Nueva carga distribuida resultante de la resta.

        Raises:
            TypeError: Si other no es una instancia de DistributedLoad.
        """
        if not isinstance(other, DistributedLoad):
            return NotImplemented
        return DistributedLoad(
            self.q_i - other.q_i,
            self.q_j - other.q_j,
            self.p_i - other.p_i,
            self.p_j - other.p_j,
            self.m_i - other.m_i,
            self.m_j - other.m_j,
            self.id or other.id
        )

    def __mul__(
        self,
        scalar: Union[float, int]
        ) -> "DistributedLoad":
        """
        Multiplica la carga por un escalar.

        Args:
            scalar: Factor de escala.

        Returns:
            DistributedLoad: Nueva carga distribuida escalada.

        Raises:
            TypeError: Si scalar no es un número.
        """
        if not isinstance(scalar, (float, int)):
            return NotImplemented
        return DistributedLoad(
            self.q_i * scalar,
            self.q_j * scalar,
            self.p_i * scalar,
            self.p_j * scalar,
            self.m_i * scalar,
            self.m_j * scalar,
            self.id
        )

    __rmul__ = __mul__

    def __truediv__(
        self,
        scalar: Union[float, int]
        ) -> "DistributedLoad":
        """
        Divide la carga por un escalar.

        Args:
            scalar: Divisor.

        Returns:
            DistributedLoad: Nueva carga distribuida dividida.

        Raises:
            TypeError: Si scalar no es un número.
            ZeroDivisionError: Si scalar es cero.
        """
        if not isinstance(scalar, (float, int)):
            return NotImplemented
        if scalar == 0:
            raise ZeroDivisionError("No se puede dividir por cero.")
        return DistributedLoad(
            self.q_i / scalar,
            self.q_j / scalar,
            self.p_i / scalar,
            self.p_j / scalar,
            self.m_i / scalar,
            self.m_j / scalar,
            self.id
        )

class PrescribedDOF:
    """
    Representa los desplazamientos prescritos en los grados de libertad de un nodo.
    """
    def __init__(
        self,
        ux: Optional[float] = None,
        uy: Optional[float] = None,
        rz: Optional[float] = None
        ) -> None:
        """
        Inicializa una nueva instancia de PrescribedDOF.

        Args:
            ux: Desplazamiento en el eje x. Por defecto es None.
            uy: Desplazamiento en el eje y. Por defecto es None.
            rz: Rotación en el eje z. Por defecto es None.
        """
        self.ux = ux
        self.uy = uy
        self.rz = rz

    def set_prescribed_dofs(
        self,
        ux: Optional[float] = None,
        uy: Optional[float] = None,
        rz: Optional[float] = None
        ) -> None:
        """
        Establece los desplazamientos prescritos en los grados de libertad del nodo.

        Args:
            ux: Desplazamiento en el eje x. Por defecto es None.
            uy: Desplazamiento en el eje y. Por defecto es None.
            rz: Rotación en el eje z. Por defecto es None.
        """
        self.ux = ux
        self.uy = uy
        self.rz = rz

    def get_prescribed_dofs(
        self
        ) -> Tuple[Optional[float], Optional[float], Optional[float]]:
        """
        Obtiene los desplazamientos prescritos en los grados de libertad del nodo.

        Returns:
            Tuple[Optional[float], Optional[float], Optional[float]]: Tupla con los desplazamientos prescritos en los grados de libertad del nodo.
        """
        return self.ux, self.uy, self.rz

    def __add__(
        self,
        other: "PrescribedDOF"
        ) -> "PrescribedDOF":
        """
        Suma dos desplazamientos prescritos.

        Args:
            other: Otra instancia de PrescribedDOF a sumar.

        Returns:
            PrescribedDOF: Nueva instancia de PrescribedDOF resultante de la suma.
        """
        if self.ux is None and other.ux is None:
            ux = None
        elif self.ux is None or other.ux is None:
            ux = self.ux or other.ux
        else:
            ux = self.ux + other.ux
        if self.uy is None and other.uy is None:
            uy = None
        elif self.uy is None or other.uy is None:
            uy = self.uy or other.uy
        else:
            uy = self.uy + other.uy
        if self.rz is None and other.rz is None:
            rz = None
        elif self.rz is None or other.rz is None:
            rz = self.rz or other.rz
        else:
            rz = self.rz + other.rz
        return PrescribedDOF(
            ux,
            uy,
            rz
        )

    def __sub__(
        self,
        other: "PrescribedDOF"
        ) -> "PrescribedDOF":
        """
        Resta dos desplazamientos prescritos.

        Args:
            other: Otra instancia de PrescribedDOF a restar.

        Returns:
            PrescribedDOF: Nueva instancia de PrescribedDOF resultante de la resta.
        """
        if self.ux is None and other.ux is None:
            ux = None
        elif self.ux is None or other.ux is None:
            ux = self.ux or other.ux
        else:
            ux = self.ux - other.ux
        if self.uy is None and other.uy is None:
            uy = None
        elif self.uy is None or other.uy is None:
            uy = self.uy or other.uy
        else:
            uy = self.uy - other.uy
        if self.rz is None and other.rz is None:
            rz = None
        elif self.rz is None or other.rz is None:
            rz = self.rz or other.rz
        else:
            rz = self.rz - other.rz
        return PrescribedDOF(
            ux,
            uy,
            rz
        )

    def __pos__(
        self
        ) -> "PrescribedDOF":
        """
        Obtiene una copia de la instancia actual.

        Returns:
            PrescribedDOF: Nueva instancia de PrescribedDOF con los mismos valores.
        """
        return PrescribedDOF(
            self.ux,
            self.uy,
            self.rz
        )

    def __neg__(
        self
        ) -> "PrescribedDOF":
        """
        Obtiene la negación de la instancia actual.

        Returns:
            PrescribedDOF: Nueva instancia de PrescribedDOF con los valores negados.
        """
        if self.ux is None:
            ux = None
        else:
            ux = -self.ux
        if self.uy is None:
            uy = None
        else:
            uy = -self.uy
        if self.rz is None:
            rz = None
        else:
            rz = -self.rz
        return PrescribedDOF(
            ux,
            uy,
            rz
        )

    def __str__(
        self
        ) -> str:
        """
        Obtiene una representación en cadena de la instancia actual.

        Returns:
            str: Representación en cadena de la instancia actual.
        """
        return f"PrescribedDOF(ux={self.ux}, uy={self.uy}, rz={self.rz})"

class ElasticSupport:
    """
    Representa los soportes elásticos en los grados de libertad de un nodo.
    """
    def __init__(
        self,
        kx: Optional[float] = None,
        ky: Optional[float] = None,
        krz: Optional[float] = None
        ) -> None:
        """
        Inicializa una nueva instancia de ElasticSupport.

        Args:
            kx: Constante de rigidez en X. Por defecto es None.
            ky: Constante de rigidez en Y. Por defecto es None.
            krz: Constante de rigidez en Z. Por defecto es None.
        """
        self.kx = kx
        self.ky = ky
        self.krz = krz

    def set_elastic_supports(
        self,
        kx: Optional[float] = None,
        ky: Optional[float] = None,
        krz: Optional[float] = None
        ) -> None:
        """
        Establece los soportes elásticos en los grados de libertad del nodo.

        Args:
            kx: Constante de rigidez en X. Por defecto es None.
            ky: Constante de rigidez en Y. Por defecto es None.
            krz: Constante de rigidez en Z. Por defecto es None.
        """
        self.kx = kx
        self.ky = ky
        self.krz = krz

    def get_elastic_supports(
        self
        ) -> Tuple[Optional[float], Optional[float], Optional[float]]:
        """
        Obtiene los soportes elásticos en los grados de libertad del nodo.

        Returns:
            Tuple[Optional[float], Optional[float], Optional[float]]: Tupla con los soportes elásticos en los grados de libertad del nodo.
        """
        return self.kx, self.ky, self.krz

    def __str__(
        self
        ) -> str:
        """
        Obtiene una representación en cadena de la instancia actual.

        Returns:
            str: Representación en cadena de la instancia actual.
        """
        return f"ElasticSupport(kx={self.kx}, ky={self.ky}, krz={self.krz})"

    def __add__(
        self,
        other: "ElasticSupport"
        ) -> "ElasticSupport":
        """
        Suma dos soportes elásticos.

        Args:
            other: Otra instancia de ElasticSupport a sumar.

        Returns:
            ElasticSupport: Nueva instancia de ElasticSupport resultante de la suma.
        """
        if self.kx is None and other.kx is None:
            kx = None
        elif self.kx is None or other.kx is None:
            kx = self.kx or other.kx
        else:
            kx = self.kx + other.kx
        if self.ky is None and other.ky is None:
            ky = None
        elif self.ky is None or other.ky is None:
            ky = self.ky or other.ky
        else:
            ky = self.ky + other.ky
        if self.krz is None and other.krz is None:
            krz = None
        elif self.krz is None or other.krz is None:
            krz = self.krz or other.krz
        else:
            krz = self.krz + other.krz
        return ElasticSupport(
            kx,
            ky,
            krz
        )

    def __sub__(
        self,
        other: "ElasticSupport"
        ) -> "ElasticSupport":
        """
        Resta dos soportes elásticos.

        Args:
            other: Otra instancia de ElasticSupport a restar.

        Returns:
            ElasticSupport: Nueva instancia de ElasticSupport resultante de la resta.
        """
        if self.kx is None and other.kx is None:
            kx = None
        elif self.kx is None or other.kx is None:
            kx = self.kx or other.kx
        else:
            kx = self.kx - other.kx
        if self.ky is None and other.ky is None:
            ky = None
        elif self.ky is None or other.ky is None:
            ky = self.ky or other.ky
        else:
            ky = self.ky - other.ky
        if self.krz is None and other.krz is None:
            krz = None
        elif self.krz is None or other.krz is None:
            krz = self.krz or other.krz
        else:
            krz = self.krz - other.krz
        return ElasticSupport(
            kx,
            ky,
            krz
        )

    def __pos__(
        self
        ) -> "ElasticSupport":
        """
        Obtiene una copia de la instancia actual.

        Returns:
            ElasticSupport: Nueva instancia de ElasticSupport con los mismos valores.
        """
        return ElasticSupport(
            self.kx,
            self.ky,
            self.krz
        )

    def __neg__(
        self
        ) -> "ElasticSupport":
        """
        Obtiene la negación de la instancia actual.

        Returns:
            ElasticSupport: Nueva instancia de ElasticSupport con los valores negados.
        """
        if self.kx is None:
            kx = None
        else:
            kx = -self.kx
        if self.ky is None:
            ky = None
        else:
            ky = -self.ky
        if self.krz is None:
            krz = None
        else:
            krz = -self.krz
        return ElasticSupport(
            kx,
            ky,
            krz
        )

class LocalAxis:
    """
    Representa el eje local de un Nodo.
    """
    def __init__(
        self,
        angle: float
        ) -> None:
        """
        Args:
            angle (float): angulo  desde  el  sistema  de  coordenadas  global  hasta  el  plano  inclinado
        """
        self.angle = angle

    def get_transformation_matrix(
        self
        ) -> np.ndarray:
        """
        Obtiene la matriz de transformación del eje local.

        Returns:
            np.ndarray: Matriz de transformación del eje local.
        """
        return np.array([
            [np.cos(self.angle), -np.sin(self.angle), 0],
            [np.sin(self.angle),  np.cos(self.angle), 0],
            [0,                   0,                  1]
        ])

class EndLengthOffset:
    """
    Representa el desplazamiento de los brazos rigidos de un elemento.
    """
    def __init__(
        self,
        la:  Optional[float]=None,
        lb:  Optional[float]=None,
        qla: Optional[bool] =None,
        qlb: Optional[bool] =None,
        fla: Optional[float]=None,
        flb: Optional[float]=None
        ) -> None:
        self.la:  Optional[float] = la    # Longitud del brazo rigido inicial
        self.lb:  Optional[float] = lb    # Longitud del brazo rigido final
        self.qla: Optional[bool]  = qla   # Indica si hay cargas en el brazo rigido inicial
        self.qlb: Optional[bool]  = qlb   # Indica si hay cargas en el brazo rigido final
        self.fla: Optional[float] = fla   # Factor de longitud del brazo rigido inicial
        self.flb: Optional[float] = flb   # Factor de longitud del brazo rigido final

class CSTLoad:
    """
    Representa la carga constante de un elemento.
    """
    def __init__(
        self,
        coordinates: np.ndarray,
        E: float,
        v: float,
        t: float,
        beta: float
        ) -> None:
        """
        Args:
            coordinates (np.ndarray): coordenadas de los nodos del elemento
            E (float): Young's modulus
            v (float): Poisson's ratio
            t (float): thickness
            beta (float): angle of the inclined plane
        """
        self.coordinates = coordinates # [[x1, y1], [x2, y2], [x3, y3]]
        self.E = E
        self.v = v
        self.t = t
        self.beta = beta

        self.Feq = np.zeros(6)

    def A(
        self
        ) -> float:
        """
        Obtiene el area del elemento.

        Returns:
            float: Area del elemento.
        """
        x1, y1 = self.coordinates[0]
        x2, y2 = self.coordinates[1]
        x3, y3 = self.coordinates[2]

        a1 = x2*y3 - x3*y2
        a2 = x3*y1 - x1*y3
        a3 = x1*y2 - x2*y1

        return (a1 + a2 + a3)/2

    def add_uniform_temperature_load(
        self,
        dt: float
        ) -> np.ndarray:
        """
        Args:
            dt (float): diferencia de temperatura
        """
        x1, y1 = self.coordinates[0]
        x2, y2 = self.coordinates[1]
        x3, y3 = self.coordinates[2]

        b1 = y2 - y3
        b2 = y3 - y1
        b3 = y1 - y2

        c1 = x3 - x2
        c2 = x1 - x3
        c3 = x2 - x1

        k = self.E * self.A() * self.t * self.beta * dt / (1 - self.v)
        Feq = k * np.array([b1, c1, b2, c2, b3, c3])

        self.Feq += Feq
        return Feq

    def add_uniform_distributed_load(
        self,
        qx: float,
        qy: float
        ) -> np.ndarray:
        """
        Añade una carga distribuida uniforme superficial en su plano del elemento
        """
        A = self.A()
        Feq = np.array([qx, qy, qx, qy, qx, qy]) * A/3 * self.t
        self.Feq += Feq
        return Feq

    def add_uniform_edge_load(
        self,
        q: float,
        edge: int
        ) -> np.ndarray:
        """
        Añade una carga distribuida uniforme en el borde del elemento
        """
        beta = np.pi - np.arctan((self.coordinates[edge][1] - self.coordinates[edge-1][1])/(self.coordinates[edge][0] - self.coordinates[edge-1][0]))
        length = np.sqrt((self.coordinates[edge][0] - self.coordinates[edge-1][0])**2 + (self.coordinates[edge][1] - self.coordinates[edge-1][1])**2)
        qx = q*np.sin(beta)
        qy = q*np.cos(beta)
        if edge == 1:
            Feq = np.array([qx, qy, qx, qy, 0, 0]) * length/2
        elif edge == 2:
            Feq = np.array([0, 0, qx, qy, qx, qy]) * length/2
        elif edge == 3:
            Feq = np.array([qx, qy, 0, 0, qx, qy]) * length/2
        self.Feq += Feq
        return Feq

    def add_linear_edge_load(
        self,
        qi: float,
        qj: float,
        edge: int
        ) -> np.ndarray:
        """
        Añade una carga distribuida lineal en el borde del elemento
        Note: el sentido es en orden de los nodos (antihorario)
        """
        beta = np.pi - np.arctan((self.coordinates[edge][1] - self.coordinates[edge-1][1])/(self.coordinates[edge][0] - self.coordinates[edge-1][0]))
        length = np.sqrt((self.coordinates[edge][0] - self.coordinates[edge-1][0])**2 + (self.coordinates[edge][1] - self.coordinates[edge-1][1])**2)
        qxi = qi*np.sin(beta)
        qyi = qi*np.cos(beta)
        qxj = qj*np.sin(beta)
        qyj = qj*np.cos(beta)
        if edge == 1:
            Feq = np.array([qxi, qyi, qxj, qyj, 0, 0]) * length/2
        elif edge == 2:
            Feq = np.array([0, 0, qxi, qyi, qxj, qyj]) * length/2
        elif edge == 3:
            Feq = np.array([qxj, qyj, 0, 0, qxi, qyi]) * length/2
        self.Feq += Feq
        return Feq



class FrameRelease:
    """
    Representa el estado de liberación de un elemento.
    """
    def __init__(
        self,
        uxi: bool=False,
        uyi: bool=False,
        rzi: bool=False,
        uxj: bool=False,
        uyj: bool=False,
        rzj: bool=False
        ) -> None:
        """
        Args:
            uxi (bool, optional): liberación en el nodo inicial en la dirección x. Defaults to False.
            uyi (bool, optional): liberación en el nodo inicial en la dirección y. Defaults to False.
            rzi (bool, optional): liberación en el nodo inicial en la dirección z. Defaults to False.
            uxj (bool, optional): liberación en el nodo final en la dirección x. Defaults to False.
            uyj (bool, optional): liberación en el nodo final en la dirección y. Defaults to False.
            rzj (bool, optional): liberación en el nodo final en la dirección z. Defaults to False.
        """
        self.uxi = uxi
        self.uyi = uyi
        self.rzi = rzi
        self.uxj = uxj
        self.uyj = uyj
        self.rzj = rzj

    def set_release(
        self, 
        uxi: bool=False, 
        uyi: bool=False, 
        rzi: bool=False, 
        uxj: bool=False, 
        uyj: bool=False, 
        rzj: bool=False
        ) -> None:
        """
        Args:
            uxi (bool, optional): liberación en el nodo inicial en la dirección x. Defaults to False.
            uyi (bool, optional): liberación en el nodo inicial en la dirección y. Defaults to False.
            rzi (bool, optional): liberación en el nodo inicial en la dirección z. Defaults to False.
            uxj (bool, optional): liberación en el nodo final en la dirección x. Defaults to False.
            uyj (bool, optional): liberación en el nodo final en la dirección y. Defaults to False.
            rzj (bool, optional): liberación en el nodo final en la dirección z. Defaults to False.
        """
        self.uxi = uxi
        self.uyi = uyi
        self.rzi = rzi
        self.uxj = uxj
        self.uyj = uyj
        self.rzj = rzj

    def get_release(
        self
        ) -> Tuple[bool, bool, bool, bool, bool, bool]:
        """
        Returns:
            Tuple[bool, bool, bool, bool, bool, bool]: liberación en el nodo inicial en la dirección x, y, z y liberación en el nodo final en la dirección x, y, z.
        """
        return self.uxi, self.uyi, self.rzi, self.uxj, self.uyj, self.rzj

    def get_dof_release(
        self
        ) -> List[int]:
        """
        Returns:
            List[int]: grados de libertad liberados.
        """
        release = []
        if self.uxi:
            release.append(0)
        if self.uyi:
            release.append(1)
        if self.rzi:
            release.append(2)
        if self.uxj:
            release.append(3)
        if self.uyj:
            release.append(4)
        if self.rzj:
            release.append(5)
        return release

    def __str__(
        self
        ) -> str:
        """
        Returns:
            str: representación en cadena de la liberación del elemento.
        """
        return f"FrameRelease(uxi={self.uxi}, uyi={self.uyi}, rzi={self.rzi}, uxj={self.uxj}, uyj={self.uyj}, rzj={self.rzj})"



from typing import TYPE_CHECKING, Dict, Optional, Union
from milcapy.material.material import Material, GenericMaterial
from milcapy.section.section import Section, RectangularSection, GenericSection, CircularSection, ShellSection
from milcapy.core.node import Node
from milcapy.core.results import Results
from milcapy.elements.member import Member
from milcapy.elements.truss import TrussElement
from milcapy.elements.CST import MembraneTriangle
from milcapy.elements.MQ6 import MembraneQuad6
from milcapy.elements.MQ6I import MembraneQuad6I
from milcapy.elements.MQ6IMod import MembraneQuad6IMod
from milcapy.elements.quad4 import MembraneQuad4
from milcapy.elements.quad8 import MembraneQuad8
from milcapy.plotter.plotter import Plotter, PlotterOptions
from milcapy.analysis.manager import AnalysisManager
from milcapy.postprocess.post_processing import PostProcessingOptions
from milcapy.loads import LoadPattern, PointLoad, PrescribedDOF, ElasticSupport, LocalAxis
from milcapy.utils.geometry import Vertex
from milcapy.utils.types import (
    CoordinateSystemType,
    StateType,
    MemberType,
    BeamTheoriesType,
    DirectionType,
    ShearCoefficientMethodType,
    LoadType,
    to_enum,
)
from milcapy.utils.types import FieldTypeMembrane, InternalForceType, ConstitutiveModel, IntegrationType, MembraneQuadElementType
import matplotlib.pyplot as plt
import matplotlib.tri as tri
import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from milcapy.utils.types import Restraints





class SystemMilcaModel:
    """
    Clase que representa el modelo estructural.
    Colecciona:
    - Materiales
    - Secciones
    - Nodos
    - Modelos de elementos finitos
    - Patrones de Carga
    - Resultados
    """

    def __init__(self) -> None:
        """
        Inicializa un nuevo modelo estructural vacío.
        """
        # Propiedades de los elementos [UNIQUE]
        self.materials: Dict[str, Material] = {}
        self.sections: Dict[str, Section] = {}

        # Elementos del modelo [UNIQUE / ADD]
        self.nodes: Dict[int, Node] = {}
        self.members: Dict[int, Member] = {}
        self.csts: Dict[int, MembraneTriangle] = {}
        self.membrane_q3dof: Dict[int, Union[MembraneQuad6, MembraneQuad6IMod]] = {}
        self.membrane_q2dof: Dict[int, Union[MembraneQuad6I, MembraneQuad4, MembraneQuad8]] = {}
        self.trusses: Dict[int, TrussElement] = {}
        # self.membrane_q6imod: Dict[int, MembraneQuad6IMod] = {} $ POR MIENTRAS SE AGREGARA A Q6
        # Patrones de carga con las asiganciones de carga en los miembros y nodos
        # {pattern_name: load pattern}
        self.load_patterns: Dict[str, LoadPattern] = {}
        self.current_load_pattern: Optional[str] = None

        # Coleccion de resultados incluyendo postprocesamiento [ADD]
        self.results: Dict[str, Results] = {}  # {pattern_name: results}

        # Analisis [UNIQUE]
        self.analysis: Optional[AnalysisManager] = None

        # Visualización [UNIQUE]
        self.plotter: Optional[Plotter] = None

        # Opciones del modelo [UNIQUE]
        self.plotter_options: "PlotterOptions" = PlotterOptions(self)
        self.postprocessing_options: "PostProcessingOptions" = PostProcessingOptions(
            factor=1, n=17)

    #! MATERIALES ##########################################################
    def add_material(
        self,
        name: str,
        modulus_elasticity: float,
        poisson_ratio: float,
        specific_weight: float = 0.0
    ) -> Material:
        """
        Agrega un material al modelo.

        Args:
            name (str): Nombre del material.
            modulus_elasticity (float): Módulo de elasticidad (E).
            poisson_ratio (float): Coeficiente de Poisson.
            specific_weight (float, opcional): Peso específico o densidad. Default es 0.0.

        Returns:
            Material: El material creado.

        Raises:
            ValueError: Si ya existe un material con el mismo nombre.
        """
        if name in self.materials:
            raise ValueError(f"Ya existe un material con el nombre '{name}'")

        material = GenericMaterial(
            name, modulus_elasticity, poisson_ratio, specific_weight)
        self.materials[name] = material
        return material

    #! SECCIONES ###########################################################
    def add_rectangular_section(
        self,
        name: str,
        material_name: str,
        base: float,
        height: float
    ) -> Section:
        """
        Agrega una sección rectangular al modelo.

        Args:
            name (str): Nombre de la sección.
            material_name (str): Nombre del material asociado (ya agregado).
            base (float): Base de la sección.
            height (float): Altura de la sección.

        Returns:
            Section: La sección creada.

        Raises:
            ValueError: Si ya existe una sección con el mismo nombre o si no existe el material.
        """
        if name in self.sections:
            raise ValueError(f"Ya existe una sección con el nombre '{name}'")

        if material_name not in self.materials:
            raise ValueError(
                f"No existe un material con el nombre '{material_name}'")

        section = RectangularSection(
            name=name,
            material=self.materials[material_name],
            base=base,
            height=height,
            shear_method=ShearCoefficientMethodType.TIMOSHENKO
        )
        self.sections[name] = section
        return section

    def add_circular_section(
        self,
        name: str,
        material_name: str,
        diameter: float,
    ) -> Section:
        """
        Agrega una sección circular al modelo.

        Args:
            name (str): Nombre de la sección.
            material_name (str): Nombre del material asociado (ya agregado).
            diameter (float): Diámetro de la sección.
        Raises:
            ValueError: Si el diámetro es menor o igual a cero.
        """
        if diameter <= 0:
            raise ValueError("El diámetro debe ser positivo.")
        if material_name not in self.materials:
            raise ValueError(
                f"No existe un material con el nombre '{material_name}'")
        section = CircularSection(
            name=name,
            material=self.materials[material_name],
            radius=diameter/2,
            shear_method=ShearCoefficientMethodType.TIMOSHENKO
        )
        self.sections[name] = section
        return section

    def add_generic_section(
        self,
        name: str,
        material_name: str,
        area: float,
        inertia: float,
        k_factor: float
    ) -> Section:
        """
        Agrega una sección rectangular al modelo.

        Args:
            name (str): Nombre de la sección.
            material_name (str): Nombre del material asociado (ya agregado).
            area (float): Área de la sección.
            inertia (float): Momento de inercia de la sección.
            k_factor (float): Coeficiente de corte de la sección.

        Returns:
            Section: La sección creada.

        Raises:
            ValueError: Si ya existe una sección con el mismo nombre o si no existe el material.
        """
        if name in self.sections:
            raise ValueError(f"Ya existe una sección con el nombre '{name}'")

        if material_name not in self.materials:
            raise ValueError(
                f"No existe un material con el nombre '{material_name}'")

        section = GenericSection(
            name=name,
            material=self.materials[material_name],
            area=area,
            inertia=inertia,
            k_factor=k_factor,
        )
        self.sections[name] = section
        return section

    def add_shell_section(
        self,
        name: str,
        material_name: str,
        thickness: float
    ) -> ShellSection:
        """
        Agrega una sección de cascaras al modelo.

        Args:
            name (str): Nombre de la sección.
            material_name (str): Nombre del material asociado (ya agregado).
            thickness (float): Grosor de la sección.

        Returns:
            ShellSection: La sección creada.

        Raises:
            ValueError: Si ya existe una sección con el mismo nombre o si no existe el material.
        """
        if name in self.sections:
            raise ValueError(f"Ya existe una sección con el nombre '{name}'")

        if material_name not in self.materials:
            raise ValueError(
                f"No existe un material con el nombre '{material_name}'")

        section = ShellSection(
            name=name,
            material=self.materials[material_name],
            thickness=thickness,
        )
        self.sections[name] = section
        return section

    def set_property_modifiers(
        self,
        section_name: str,
        axial_area: Optional[float] = 1,
        shear_area: Optional[float] = 1,
        moment_inertia: Optional[float] = 1,
        weight: Optional[float] = 1
    ) -> None:
        """
        Establece los modificadores de propiedad de una sección.

        Args:
            section_name (str): Nombre de la sección.
            axial_area (float): Modificador de área transversal.
            shear_area (float): Modificador de área de corte.
            moment_inertia (float): Modificador de momento de inercia.
            weight (float): Modificador de peso.
        """
        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        section = self.sections[section_name]
        section.kA = axial_area
        section.kAs = shear_area
        section.kI = moment_inertia
        section.kg = weight

    #! NODOS ###############################################################
    def add_node(
        self,
        id: int,
        x: float,
        y: float
    ) -> Node:
        """
        Agrega un nodo al modelo.

        Args:
            id (int): Identificador del nodo.
            x (float): Coordenada x del nodo.
            y (float): Coordenada y del nodo.

        Returns:
            Node: El nodo creado.

        Raises:
            ValueError: Si ya existe un nodo con el mismo ID.
        """
        if id in self.nodes:
            raise ValueError(f"Ya existe un nodo con el ID {id}")

        node = Node(id, Vertex(x, y))
        self.nodes[id] = node
        return node

    #! ELEMENTOS ###########################################################
    def add_member(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        section_name: str,
        beam_theory: Union[str,
                           BeamTheoriesType] = BeamTheoriesType.TIMOSHENKO,
        # member_type: Union[str, MemberType] = MemberType.FRAME
    ) -> Member:
        """
        Agrega un miembro estructural al modelo, por defecto usa la teoría de Timoshenko para vigas.

        Args:
            id (int): Identificador del miembro.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            beam_theory (str, opcional): Teoría de la viga. Por defecto es Timoshenko.

        Returns:
            Member: El miembro creado.

        Raises:
            ValueError: Si ya existe un miembro con el mismo ID, o si no existen los nodos o la sección.
        """
        self.__id_verifier(id, node_i_id, node_j_id)

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        # # Convertir a enum si es string
        # if isinstance(member_type, str):
        #     member_type = to_enum(member_type, MemberType)
        member_type = MemberType.FRAME

        if isinstance(beam_theory, str):
            beam_theory = to_enum(beam_theory, BeamTheoriesType)

        element = Member(
            id=id,
            node_i=self.nodes[node_i_id],
            node_j=self.nodes[node_j_id],
            section=self.sections[section_name],
            member_type=member_type,
            beam_theory=beam_theory,
        )
        self.members[id] = element
        return element

    def add_elastic_timoshenko_beam(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        section_name: str,
    ) -> Member:
        """
        Agrega una viga de Timoshenko al modelo con rigidez axial.

        Args:
            id (int): Identificador del miembro.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.

        Returns:
            Member: El miembro creado.

        Raises:
            ValueError: Si ya existe un elemento con el mismo ID, o si no existen los nodos o la sección.
        """
        self.__id_verifier(id, node_i_id, node_j_id)

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        beam_theory = BeamTheoriesType.TIMOSHENKO

        element = Member(
            id=id,
            node_i=self.nodes[node_i_id],
            node_j=self.nodes[node_j_id],
            section=self.sections[section_name],
            member_type=MemberType.FRAME,
            beam_theory=beam_theory,
        )
        self.members[id] = element
        return element

    def add_elastic_euler_bernoulli_beam(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        section_name: str,
    ) -> Member:
        """
        Agrega una viga de Euler-Bernoulli al modelo con rigidez axial.

        Args:
            id (int): Identificador del miembro.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.

        Returns:
            Member: El miembro creado.

        Raises:
            ValueError: Si ya existe un elemento con el mismo ID, o si no existen los nodos o la sección.
        """
        self.__id_verifier(id, node_i_id, node_j_id)

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        beam_theory = BeamTheoriesType.EULER_BERNOULLI

        element = Member(
            id=id,
            node_i=self.nodes[node_i_id],
            node_j=self.nodes[node_j_id],
            section=self.sections[section_name],
            member_type=MemberType.FRAME,
            beam_theory=beam_theory,
        )
        self.members[id] = element
        return element

    def add_truss(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        section_name: str
    ) -> TrussElement:
        """
        Agrega un elemento de trusse al modelo.

        Args:
            id (int): Identificador del elemento.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.

        Returns:
            TrussElement: El elemento creado.

        Raises:
            ValueError: Si ya existe un elemento con el mismo ID, o si no existen los nodos o la sección.
        """
        self.__id_verifier(id, node_i_id, node_j_id)

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        TE = TrussElement(
            id=id,
            node_i=self.nodes[node_i_id],
            node_j=self.nodes[node_j_id],
            section=self.sections[section_name],
        )
        self.trusses[id] = TE
        return TE


    #! ELEMENTOS FINITOS ###################################################
    def add_cst(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        node_k_id: int,
        section_name: str,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
    ) -> MembraneTriangle:
        """
        Agrega un triángulo de membrana al modelo.

        Args:
            id (int): Identificador del triángulo.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            node_k_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            state (Union[ConstitutiveModel, str]): Estado constitutivo del triángulo.

        Returns:
            MembraneTriangle: El triángulo creado.

        Raises:
            ValueError: Si ya existe un elemento con el mismo ID, o si no existen los nodos o el material.
            ValueError: Si el estado constitutivo no es 'PLANE_STRESS' o 'PLANE_STRAIN'.
            ValueError: Si los nodos no forman un polígono en sentido antihorario.

        Notes:
            El elemento agregado es un CONSTANT STRAIN TRIANGLE (CST).
            La enumeracion (orden) de los nodos debe ser en sentido antihorario.
        """
        self.__id_verifier(id, node_i_id, node_j_id, node_k_id)

        if self.__is_counterclockwise(node_i_id, node_j_id, node_k_id):
            raise ValueError("Los nodos no forman un triángulo en sentido antihorario")

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        if isinstance(state, str):
            state = to_enum(state, ConstitutiveModel)
        if state not in [ConstitutiveModel.PLANE_STRESS, ConstitutiveModel.PLANE_STRAIN]:
            raise ValueError(
                f"El estado constitutivo debe ser 'PLANE_STRESS' o 'PLANE_STRAIN'")

        cst = MembraneTriangle(
            id=id,
            node1=self.nodes[node_i_id],
            node2=self.nodes[node_j_id],
            node3=self.nodes[node_k_id],
            section=self.sections[section_name],
            state=state,
        )
        self.csts[id] = cst
        return cst

    def add_membrane_q6(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        node_k_id: int,
        node_l_id: int,
        section_name: str,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
    ) -> MembraneQuad6:
        """
        Agrega un cuadrilátero de membrana de 6 nodos al modelo con rigidez a la perforación (3dof/nodo).

        Args:
            id (int): Identificador del cuadrilátero.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            node_k_id (int): ID del nodo final.
            node_l_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            state (Union[ConstitutiveModel, str]): Estado constitutivo del cuadrilátero.

        Returns:
            MembraneQuad6: El cuadrilátero creado.

        Raises:
            ValueError: Si ya existe un cuadrilátero con el mismo ID, o si no existen los nodos o la sección.
            ValueError: Si el estado constitutivo no es 'PLANE_STRESS' o 'PLANE_STRAIN'.
            ValueError: Si los nodos no forman un cuadrilátero en sentido antihorario.

        Notes:
            El elemento agregado es un Q6 el mismo elemento que platea el dr. wilson en su libro "analisis estatico y dinamico de estructuras" membranas con grados de libertad de perforación.
            La enumeracion (orden) de los nodos debe ser en sentido antihorario.
        """
        self.__id_verifier(id, node_i_id, node_j_id, node_k_id, node_l_id)

        if self.__is_counterclockwise(node_i_id, node_j_id, node_k_id, node_l_id):
            raise ValueError("Los nodos no forman un cuadrilátero en sentido antihorario")

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        if isinstance(state, str):
            state = to_enum(state, ConstitutiveModel)
        if state not in [ConstitutiveModel.PLANE_STRESS, ConstitutiveModel.PLANE_STRAIN]:
            raise ValueError(
                f"El estado constitutivo debe ser 'PLANE_STRESS' o 'PLANE_STRAIN'")

        MQ6 = MembraneQuad6(
            id=id,
            node1=self.nodes[node_i_id],
            node2=self.nodes[node_j_id],
            node3=self.nodes[node_k_id],
            node4=self.nodes[node_l_id],
            section=self.sections[section_name],
            state=state,
        )
        self.membrane_q3dof[id] = MQ6
        return MQ6

    def add_membrane_q6i(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        node_k_id: int,
        node_l_id: int,
        section_name: str,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
    ) -> MembraneQuad6I:
        """
        Agrega un rectangular de membrana de 4 nodos y con modos incompatibles al modelo, sin rigidez a la perforación.

        Args:
            id (int): Identificador del cuadrilátero.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            node_k_id (int): ID del nodo final.
            node_l_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            state (Union[ConstitutiveModel, str]): Estado constitutivo del cuadrilátero.

        Returns:
            MembraneQuad6I: El cuadrilátero creado.

        Raises:
            ValueError: Si ya existe un cuadrilátero con el mismo ID, o si no existen los nodos o la sección.
            ValueError: Si el estado constitutivo no es 'PLANE_STRESS' o 'PLANE_STRAIN'.
            ValueError: Si los nodos no forman un cuadrilátero en sentido antihorario.

        Notes:
            El elemento agregado es un Q4I el mismo modelo que usa ETABS, SAP2000.
            La enumeracion (orden) de los nodos debe ser en sentido antihorario.
        """

        self.__id_verifier(id, node_i_id, node_j_id, node_k_id, node_l_id)

        # # verificar si es rectangular
        # if not self.__is_rectangular(node_i_id, node_j_id, node_k_id, node_l_id):
        #     raise ValueError(
        #         "Los nodos no forman un rectángulo en sentido antihorario")

        if self.__is_counterclockwise(node_i_id, node_j_id, node_k_id, node_l_id):
            raise ValueError(
                "Los nodos no forman un rectángulo en sentido antihorario")

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        if isinstance(state, str):
            state = to_enum(state, ConstitutiveModel)
        if state not in [ConstitutiveModel.PLANE_STRESS, ConstitutiveModel.PLANE_STRAIN]:
            raise ValueError(
                f"El estado constitutivo debe ser 'PLANE_STRESS' o 'PLANE_STRAIN'")

        MQ6I = MembraneQuad6I(
            id=id,
            node1=self.nodes[node_i_id],
            node2=self.nodes[node_j_id],
            node3=self.nodes[node_k_id],
            node4=self.nodes[node_l_id],
            section=self.sections[section_name],
            state=state,
        )
        self.membrane_q2dof[id] = MQ6I
        return MQ6I

    def add_membrane_q6i_mod(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        node_k_id: int,
        node_l_id: int,
        section_name: str,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
        ele_type: Union[MembraneQuadElementType, str] = MembraneQuadElementType.MQ6I,
    ) -> MembraneQuad6IMod:
        """
        Agrega un cuadrilátero de membrana de 4 nodos y con modos incompatibles al modelo, con rigidez a la perforación.

        Args:
            id (int): Identificador del cuadrilátero.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            node_k_id (int): ID del nodo final.
            node_l_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            state (Union[ConstitutiveModel, str]): Estado constitutivo del cuadrilátero.
            ele_type (Union[MembraneQuadElementType, str]): Tipo de elemento de membrana.

        Returns:
            MembraneQuad6IMod: El cuadrilátero creado.

        Raises:
            ValueError: Si ya existe un cuadrilátero con el mismo ID, o si no existen los nodos o la sección.
            ValueError: Si los nodos no forman un cuadrilátero en sentido antihorario.
            ValueError: Si los nodos no forman un rectángulo.

        Notes:
            El elemento agregado es un QUAD6I el mismo modelo que usa ETABS, SAP2000.
            La enumeracion (orden) de los nodos debe ser en sentido antihorario.
            OJO: es un modelo rigidez combinada no esta documentado en ningun libro (no confiar)
        """

        self.__id_verifier(id, node_i_id, node_j_id, node_k_id, node_l_id)
        # if not self.__is_counterclockwise(node_i_id, node_j_id, node_k_id, node_l_id):
        #     raise ValueError(
        #         "Los nodos no forman un rectángulo en sentido antihorario")

        # if not self.__is_rectangular(node_i_id, node_j_id, node_k_id, node_l_id):
        #     raise ValueError(
        #         "Los nodos no forman un rectángulo")

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        if isinstance(state, str):
            state = to_enum(state, ConstitutiveModel)
        if state not in [ConstitutiveModel.PLANE_STRESS, ConstitutiveModel.PLANE_STRAIN]:
            raise ValueError(
                f"El estado constitutivo debe ser 'PLANE_STRESS' o 'PLANE_STRAIN'")

        if isinstance(ele_type, str):
            ele_type = to_enum(ele_type, MembraneQuadElementType)
        if ele_type not in [MembraneQuadElementType.MQ4, MembraneQuadElementType.MQ6, MembraneQuadElementType.MQ6I, MembraneQuadElementType.MQ8Reduced, MembraneQuadElementType.MQ8Complete]:
            raise ValueError(
                f"El tipo de elemento debe ser 'MQ4', 'MQ6', 'MQ6I', 'MQ8Reduced' o 'MQ8Complete'")

        MQ6IMod = MembraneQuad6IMod(
            id=id,
            node1=self.nodes[node_i_id],
            node2=self.nodes[node_j_id],
            node3=self.nodes[node_k_id],
            node4=self.nodes[node_l_id],
            section=self.sections[section_name],
            state=state,
            ele_type=ele_type,
        )
        self.membrane_q3dof[id] = MQ6IMod
        return MQ6IMod

    def add_membrane_q4(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        node_k_id: int,
        node_l_id: int,
        section_name: str,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
    ) -> MembraneQuad4:
        """
        Agrega un cuadrilátero de membrana de 4 nodos al modelo (2dof/nodo).

        Args:
            id (int): Identificador del cuadrilátero.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            node_k_id (int): ID del nodo final.
            node_l_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            state (Union[ConstitutiveModel, str], optional): Estado constitutivo del cuadrilátero. Defaults to ConstitutiveModel.PLANE_STRESS.

        Raises:
            ValueError: Si ya existe un cuadrilátero con el mismo ID, o si no existen los nodos o la sección.
            ValueError: Si los nodos no forman un cuadrilátero en sentido antihorario.
            ValueError: Si los nodos no forman un rectángulo.

        Returns:
            MembraneQuad4: El cuadrilátero creado.

        Notes:
            El elemento agregado es un Q4.
            La enumeracion (orden) de los nodos debe ser en sentido antihorario.
            Tiene muy poca convergencia, recomiendo usar Q8.
        """
        self.__id_verifier(id, node_i_id, node_j_id, node_k_id, node_l_id)

        if self.__is_counterclockwise(node_i_id, node_j_id, node_k_id, node_l_id):
            raise ValueError("Los nodos no forman un cuadrilátero en sentido antihorario")

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        if isinstance(state, str):
            state = to_enum(state, ConstitutiveModel)
        if state not in [ConstitutiveModel.PLANE_STRESS, ConstitutiveModel.PLANE_STRAIN]:
            raise ValueError(
                f"El estado constitutivo debe ser 'PLANE_STRESS' o 'PLANE_STRAIN'")

        MQ4 = MembraneQuad4(
            id=id,
            node1=self.nodes[node_i_id],
            node2=self.nodes[node_j_id],
            node3=self.nodes[node_k_id],
            node4=self.nodes[node_l_id],
            section=self.sections[section_name],
            state=state,
        )
        self.membrane_q2dof[id] = MQ4
        return MQ4

    def add_membrane_q8(
        self,
        id: int,
        node_i_id: int,
        node_j_id: int,
        node_k_id: int,
        node_l_id: int,
        section_name: str,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
        integration: Union[IntegrationType, str] = IntegrationType.COMPLETE,
    ) -> MembraneQuad8:
        """
        Agrega un cuadrilátero de membrana de 8 nodos al modelo con rigidez a la perforación (2dof/nodo).

        Args:
            id (int): Identificador del cuadrilátero.
            node_i_id (int): ID del nodo inicial.
            node_j_id (int): ID del nodo final.
            node_k_id (int): ID del nodo final.
            node_l_id (int): ID del nodo final.
            section_name (str): Nombre de la sección asociada.
            state (Union[ConstitutiveModel, str]): Estado constitutivo del cuadrilátero.
            integration (Union[IntegrationType, str]): Tipo de integración.

        Returns:
            MembraneQuad6: El cuadrilátero creado.

        Raises:
            ValueError: Si ya existe un cuadrilátero con el mismo ID, o si no existen los nodos o la sección.
            ValueError: Si el estado constitutivo no es 'PLANE_STRESS' o 'PLANE_STRAIN'.
            ValueError: Si los nodos no forman un cuadrilátero en sentido antihorario.

        Notes:
            La enumeracion (orden) de los nodos debe ser en sentido antihorario.
        """
        self.__id_verifier(id, node_i_id, node_j_id, node_k_id, node_l_id)

        if self.__is_counterclockwise(node_i_id, node_j_id, node_k_id, node_l_id):
            raise ValueError("Los nodos no forman un cuadrilátero en sentido antihorario")

        if section_name not in self.sections:
            raise ValueError(
                f"No existe una sección con el nombre '{section_name}'")

        if isinstance(state, str):
            state = to_enum(state, ConstitutiveModel)
        if state not in [ConstitutiveModel.PLANE_STRESS, ConstitutiveModel.PLANE_STRAIN]:
            raise ValueError(
                f"El estado constitutivo debe ser 'PLANE_STRESS' o 'PLANE_STRAIN'")

        if isinstance(integration, str):
            integration = to_enum(integration, IntegrationType)
        if integration not in [IntegrationType.COMPLETE, IntegrationType.REDUCED]:
            raise ValueError(
                f"El tipo de integración debe ser 'COMPLETE' o 'REDUCED'")

        MQ8 = MembraneQuad8(
            id=id,
            node1=self.nodes[node_i_id],
            node2=self.nodes[node_j_id],
            node3=self.nodes[node_k_id],
            node4=self.nodes[node_l_id],
            section=self.sections[section_name],
            state=state,
            integration=integration,
        )
        self.membrane_q2dof[id] = MQ8
        return MQ8

    #! CONDICIONES DE FRONTERA #############################################

    def add_restraint(
        self,
        node_id: int,
        ux: bool,
        uy: bool,
        rz: bool,
    ) -> None:
        """
        Asigna restricciones (condiciones de frontera) a un nodo.

        Args:
            node_id (int): Identificador del nodo.
            ux (bool): Restricción de traslación en el eje x.
            uy (bool): Restricción de traslación en el eje y.
            rz (bool): Restricción de rotación en el eje z.

        Raises:
            ValueError: Si no existe el nodo.
        """
        if node_id not in self.nodes:
            raise ValueError(f"No existe un nodo con el ID {node_id}")
        if not all(isinstance(x, bool) for x in [ux, uy, rz]):
            raise ValueError(
                "ux, uy y rz deben ser booleanos")

        self.nodes[node_id].set_restraints((ux, uy, rz))


    #! PATRONES DE CARGA ###################################################
    def add_load_pattern(
        self,
        name: str,
        self_weight_multiplier: float = 0.0,
        state: Union[str, StateType] = StateType.ACTIVE
    ) -> LoadPattern:
        """
        Agrega un patrón de carga al modelo.

        Args:
            name (str): Nombre del patrón de carga.
            self_weight_multiplier (float, opcional): Multiplicador del peso propio. Default es 0.0.
            state (str o State, opcional): Estado del patrón. Default es ACTIVE.

        Returns:
            LoadPattern: El patrón de carga creado.

        Raises:
            ValueError: Si ya existe un patrón con el mismo nombre.
        """
        if name in self.load_patterns:
            raise ValueError(
                f"Ya existe un patrón de carga con el nombre '{name}'")

        if isinstance(state, str):
            state = to_enum(state, StateType)

        load_pattern = LoadPattern(
            name=name,
            self_weight_multiplier=self_weight_multiplier,
            state=state,
            system=self,
        )
        self.load_patterns[name] = load_pattern
        return load_pattern

    def set_state_load_pattern(
        self,
        load_pattern_name: str,
        state: Union[str, StateType]
    ) -> None:
        """
        Cambia el estado de un patrón de carga.

        Args:
            load_pattern_name (str): Nombre del patrón de carga.
            state (str o State, opcional): Estado del patrón. Default es ACTIVE.

        Raises:
            ValueError: Si no existe el patrón de carga.
        """
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")

        if isinstance(state, str):
            state = to_enum(state, StateType)

        self.load_patterns[load_pattern_name].state = state

    #! CARGAS ##############################################################
    def add_point_load(
        self,
        node_id: int,
        load_pattern_name: str,
        fx: float = 0.0,
        fy: float = 0.0,
        mz: float = 0.0,
        CSys: Union[str, CoordinateSystemType] = "GLOBAL",
        replace: bool = False
    ) -> None:
        """
        Asigna una carga puntual a un nodo dentro de un patrón de carga.

        Args:
            node_id (int): Identificador del nodo.
            load_pattern_name (str): Nombre del patrón de carga.
            CSys (str o CoordinateSystemType, opcional): Sistema de coordenadas. Default es "GLOBAL".
            fx (float, opcional): Fuerza en X. Default es 0.0.
            fy (float, opcional): Fuerza en Y. Default es 0.0.
            mz (float, opcional): Momento en Z. Default es 0.0.
            angle_rot (float, opcional): Ángulo de rotación en radianes. Default es None.
            replace (bool, opcional): Si se reemplaza la carga existente. Default es False.

        Raises:
            ValueError: Si no existe el nodo o el patrón de carga.
        """
        if node_id not in self.nodes:
            raise ValueError(f"No existe un nodo con el ID {node_id}")

        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")

        # Convertir a enum si es string
        if isinstance(CSys, str):
            csys_enum = to_enum(CSys, CoordinateSystemType)
        else:
            csys_enum = CSys

        if csys_enum == CoordinateSystemType.LOCAL:
            angle_rot = self.nodes[node_id].local_axis.angle
        else:
            angle_rot = None

        self.load_patterns[load_pattern_name].add_point_load(
            node_id=node_id,
            forces=PointLoad(fx=fx, fy=fy, mz=mz),
            csys=csys_enum,
            angle_rot=angle_rot,
            replace=replace
        )

    def add_distributed_load(
        self,
        member_id: int,
        load_pattern_name: str,
        load_start: float = 0.0,
        load_end: float = 0.0,
        CSys: Union[str, CoordinateSystemType] = "LOCAL",
        direction: Union[str, DirectionType] = "LOCAL_2",
        load_type: Union[str, LoadType] = "FORCE",
        replace: bool = False,
    ) -> None:
        """
        Asigna una carga distribuida a un miembro dentro de un patrón de carga.

        Args:
            member_id (int): Identificador del miembro.
            load_pattern_name (str): Nombre del patrón de carga.
            load_start (float, opcional): Magnitud de la carga en el inicio. Default es 0.0.
            load_end (float, opcional): Magnitud de la carga en el final. Default es 0.0.
            CSys (str o CoordinateSystemType, opcional): Sistema de coordenadas. Default es "LOCAL".
            direction (str o DirectionType, opcional): Dirección de la carga. Default es "LOCAL_2".
            load_type (str o LoadType, opcional): Tipo de carga. Default es "FORCE".
            replace (bool, opcional): Si se reemplaza la carga existente. Default es False.

        Raises:
            ValueError: Si no existe el miembro o el patrón de carga.
        """
        if member_id not in list(self.members.keys()) + list(self.trusses.keys()):
            raise ValueError(f"No existe un miembro con el ID {member_id}")

        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")

        # Convertir a enum si son strings
        if isinstance(CSys, str):
            csys_enum = to_enum(CSys, CoordinateSystemType)
        else:
            csys_enum = CSys

        if isinstance(direction, str):
            direction_enum = to_enum(direction, DirectionType)
        else:
            direction_enum = direction

        if isinstance(load_type, str):
            load_type_enum = to_enum(load_type, LoadType)
        else:
            load_type_enum = load_type

        if isinstance(self.trusses.get(member_id), TrussElement) and direction_enum != DirectionType.LOCAL_1:
            raise ValueError("Solo se puede aplicar carga distribuida axial en Armaduras osea solo en la direccion LOCAL_1")

        self.load_patterns[load_pattern_name].add_distributed_load(
            member_id=member_id,
            load_start=load_start,
            load_end=load_end,
            csys=csys_enum,
            direction=direction_enum,
            load_type=load_type_enum,
            replace=replace,
        )

    def add_self_weight(
        self,
        load_pattern_name: str,
        factor: float = 1
    ) -> None:
        """
        Asigna el peso propio a todos los miembros del modelo dentro de un patrón de carga.

        Args:
            load_pattern_name (str): Nombre del patrón de carga.
            factor (float, opcional): Factor de escala para el peso propio. Default es 1.

        Raises:
            ValueError: Si no existe el patrón de carga.
        """
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")
        for member in self.members.values():
            self.add_distributed_load(
                member_id=member.id,
                load_pattern_name=load_pattern_name,
                load_start=member.section.A()/member.section.kA * member.section.g() * factor,
                load_end=member.section.A()/member.section.kA * member.section.g() * factor,
                CSys="GLOBAL",
                direction="GRAVITY",
                load_type="FORCE",
            )

    def add_cst_uniform_temperature_load(
            self,
            cst_id: int,
            load_pattern_name: str,
            dt: float) -> None:
        """
        Asigna una carga uniforme de temperatura a un CST dentro de un patrón de carga.

        Args:
            cst_id (int): ID del CST.
            load_pattern_name (str): Nombre del patrón de carga.
            dt (float): Incremento de temperatura.

        Raises:
            ValueError: Si no existe el CST o el patrón de carga.
        """
        if cst_id not in self.csts:
            raise ValueError(f"No existe un miembro con el ID {cst_id}")
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")

        self.load_patterns[load_pattern_name].add_cst_uniform_temperature_load(
            cst_id=cst_id,
            dt=dt
        )

    def add_cst_uniform_distributed_load(
        self,
        cst_id: int,
        load_pattern_name: str,
        qx: float,
        qy: float,
    ) -> None:
        """
        Asigna una carga distribuida uniforme a un CST dentro de un patrón de carga.

        Args:
            cst_id (int): ID del CST.
            load_pattern_name (str): Nombre del patrón de carga.
            qx (float): Carga distribuida en X.
            qy (float): Carga distribuida en Y.

        Raises:
            ValueError: Si no existe el CST o el patrón de carga.
        """
        if cst_id not in self.csts:
            raise ValueError(f"No existe un miembro con el ID {cst_id}")
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")

        self.load_patterns[load_pattern_name].add_cst_uniform_distributed_load(
            cst_id=cst_id,
            qx=qx,
            qy=qy
        )

    def add_cst_uniform_edge_load(
        self,
        cst_id: int,
        load_pattern_name: str,
        q: float,
        edge: int
    ) -> None:
        """
        Asigna una carga uniforme a una cara de un CST dentro de un patrón de carga.

        Args:
            cst_id (int): ID del CST.
            load_pattern_name (str): Nombre del patrón de carga.
            q (float): Carga uniforme.
            edge (int): Cara del CST (1, 2, o 3).

        Raises:
            ValueError: Si no existe el CST o el patrón de carga, o si la cara no es válida.
        """
        if cst_id not in self.csts:
            raise ValueError(f"No existe un miembro con el ID {cst_id}")
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")
        if edge not in [1, 2, 3]:
            raise ValueError(f"Las caras(edge) solo puenden ser [1, 2, 3]")

        self.load_patterns[load_pattern_name].add_cst_uniform_edge_load(
            cst_id=cst_id,
            q=q,
            edge=edge
        )

    def add_cst_linear_edge_load(
        self,
        cst_id: int,
        load_pattern_name: str,
        qi: float,
        qj: float,
        edge: int
    ) -> None:
        """
        Asigna una carga lineal a una cara de un CST dentro de un patrón de carga.

        Args:
            cst_id (int): ID del CST.
            load_pattern_name (str): Nombre del patrón de carga.
            qi (float): Carga en el inicio.
            qj (float): Carga en el final.
            edge (int): Cara del CST (1, 2, o 3).

        Raises:
            ValueError: Si no existe el CST o el patrón de carga, o si la cara no es válida.
        """
        if cst_id not in self.csts:
            raise ValueError(f"No existe un miembro con el ID {cst_id}")
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")
        if edge not in [1, 2, 3]:
            raise ValueError(f"Las caras(edge) solo puenden ser [1, 2, 3]")

        self.load_patterns[load_pattern_name].add_cst_uniform_edge_load(
            cst_id=cst_id,
            qi=qi,
            qj=qj,
            edge=edge
        )


    #! TOPICO ESPECIALES DE ANALISIS #######################################
    def add_end_length_offset(
        self,
        member_id: int,
        la: float = 0,
        lb: float = 0,
        qla: bool = True,
        qlb: bool = True,
        fla: float = 1,
        flb: float = 1
    ) -> None:
        """
        Asigna un desplazamiento de longitud final a un miembro dentro de un patrón de carga.

        Args:
            member_id (int): ID del miembro.
            la (float): Desface de longitud final en el nodo inicial.
            lb (float): Desface de longitud final en el nodo final.
            qla (bool, opcional): Si se aplica el carga en el brazo inicial. Default es True.
            qlb (bool, opcional): Si se aplica el carga en el brazo final. Default es True.
            fla (float, opcional): Factor de zona rigida del brazo inicial. Default es 1.
            flb (float, opcional): Factor de zona rigida del brazo final. Default es 1.

        Raises:
            ValueError: Si no existe el miembro.
        """
        if member_id not in self.members:
            raise ValueError(f"No existe un miembro con el ID {member_id}")
        self.members[member_id].la = la
        self.members[member_id].lb = lb
        self.members[member_id].qla = qla
        self.members[member_id].qlb = qlb
        self.members[member_id].fla = fla
        self.members[member_id].flb = flb

    def add_prescribed_dof(
        self,
        node_id: int,
        load_pattern_name: str,
        ux: Optional[float] = None,
        uy: Optional[float] = None,
        rz: Optional[float] = None,
        CSys: Union[str, CoordinateSystemType] = "GLOBAL"
    ) -> None:
        """
        Asigna un desplazamiento prescrito a un nodo dentro de un patrón de carga.

        Args:
            node_id (int): Identificador del nodo.
            load_pattern_name (str): Nombre del patrón de carga.
            ux (float, opcional): Desplazamiento en X. Default es None.
            uy (float, opcional): Desplazamiento en Y. Default es None.
            rz (float, opcional): Desplazamiento en Z. Default es None.
            CSys (str, opcional): Sistema de coordenadas. Default es "GLOBAL".
        Raises:
            ValueError: Si no existe el nodo o el patrón de carga.
        """
        if node_id not in self.nodes:
            raise ValueError(f"No existe un nodo con el ID {node_id}")
        if load_pattern_name not in self.load_patterns:
            raise ValueError(
                f"No existe un patrón de carga con el nombre '{load_pattern_name}'")
        if isinstance(CSys, str):
            CSys = to_enum(CSys, CoordinateSystemType)
        if CSys not in [CoordinateSystemType.LOCAL, CoordinateSystemType.GLOBAL]:
            raise ValueError(f"CSys debe ser 'LOCAL' o 'GLOBAL'")
        if CSys == CoordinateSystemType.GLOBAL and self.nodes[node_id].local_axis is not None:
            angle = self.nodes[node_id].local_axis.angle
            T = np.array([
                [np.cos(angle), np.sin(angle), 0],
                [-np.sin(angle), np.cos(angle), 0],
                [0, 0, 1]
            ]).T
            ux, uy, rz = np.dot(T, np.array([ux or 0, uy or 0, rz or 0]))
            ux = ux if ux != 0 else None
            uy = uy if uy != 0 else None
            rz = rz if rz != 0 else None
        self.load_patterns[load_pattern_name].add_prescribed_dof(
            node_id, PrescribedDOF(ux=ux, uy=uy, rz=rz))

    def add_elastic_support(
        self,
        node_id: int,
        kx: Optional[float] = None,
        ky: Optional[float] = None,
        krz: Optional[float] = None,
        CSys: Union[str, CoordinateSystemType] = "GLOBAL"
    ) -> None:
        """
        Asigna un apoyo elástico a un nodo.

        Args:
            node_id (int): Identificador del nodo.
            kx (float, opcional): Constante de rigidez en X. Default es None.
            ky (float, opcional): Constante de rigidez en Y. Default es None.
            krz (float, opcional): Constante de rigidez en Z. Default es None.
            CSys (str, opcional): Sistema de coordenadas. Default es "GLOBAL".
        Raises:
            ValueError: Si no existe el nodo.
        """
        if node_id not in self.nodes:
            raise ValueError(f"No existe un nodo con el ID {node_id}")
        if isinstance(CSys, str):
            CSys = to_enum(CSys, CoordinateSystemType)
        if CSys not in [CoordinateSystemType.LOCAL, CoordinateSystemType.GLOBAL]:
            raise ValueError(f"CSys debe ser 'LOCAL' o 'GLOBAL'")
        if CSys == CoordinateSystemType.GLOBAL and self.nodes[node_id].local_axis is not None:
            angle = self.nodes[node_id].local_axis.angle
            T = np.array([
                [np.cos(angle), np.sin(angle), 0],
                [-np.sin(angle), np.cos(angle), 0],
                [0, 0, 1]
            ])
            kx, ky, krz = np.dot(T, np.array([kx or 0, ky or 0, krz or 0]))
            kx = kx if kx != 0 else None
            ky = ky if ky != 0 else None
            krz = krz if krz != 0 else None
        self.nodes[node_id].set_elastic_support(
            ElasticSupport(kx=kx, ky=ky, krz=krz))

    def add_local_axis_for_node(
        self,
        node_id: int,
        angle: float
    ) -> None:
        """
        Asigna un eje local a un nodo.
        >> con esto se puede hacer soportes inclinados y obtener resultados en el eje local del NODO
        Args:
            node_id (int): Identificador del nodo.
            angle (float): Ángulo del eje local en grados.

        Raises:
            ValueError: Si no existe el nodo.
        """
        if node_id not in self.nodes:
            raise ValueError(f"No existe un nodo con el ID {node_id}")
        local_axis = LocalAxis(angle)
        self.nodes[node_id].set_local_axis(local_axis)

    def add_releases(
        self,
        member_id: int,
        pi: bool = False,
        vi: bool = False,
        mi: bool = False,
        pj: bool = False,
        vj: bool = False,
        mj: bool = False
    ):
        """
        Asigna liberaciones a un miembro.

        Args:
            member_id (int): Identificador del miembro.
            pi (bool, opcional): Liberación de fuerza axial del nodo inicial. Default es False.
            vi (bool, opcional): Liberación de cortante del nodo inicial. Default es False.
            mi (bool, opcional): Liberación de momento del nodo inicial. Default es False.
            pj (bool, opcional): Liberación de fuerza axial del nodo final. Default es False.
            vj (bool, opcional): Liberación de cortante del nodo final. Default es False.
            mj (bool, opcional): Liberación de momento del nodo final. Default es False.

        Raises:
            ValueError: Si no existe el miembro.
        """
        if member_id not in self.members:
            raise ValueError(f"No existe un miembro con el ID {member_id}")

        # Reglas de exclusión
        if pi and pj:
            raise ValueError(
                "Liberación inválida: no se pueden liberar P en ambos extremos (pi y pj).")
        if vi and vj:
            raise ValueError(
                "Liberación inválida: no se pueden liberar V en ambos extremos (vi y vj).")

        self.members[member_id].add_releases(
            uxi=pi, uyi=vi, rzi=mi, uxj=pj, uyj=vj, rzj=mj)

    #! ANALISIS ##############################################################
    def solve(
        self,
        load_pattern_name: list[str] | None = None
    ) -> Dict[str, Results]:
        """
        Resuelve el sistema estructural aplicando el método de rigidez:
        - Asigna las cargas a nodos y elementos.
        - Calcula el vector de fuerzas y la matriz de rigidez global.
        - Resuelve el sistema de ecuaciones para obtener los desplazamientos y reacciones.

        Returns:
            Dict[str, Results]: Objeto con los resultados del análisis.

        Raises:
            ValueError: Si no hay patrones de carga definidos.
        """
        # Verificar que exista al menos un patrón de carga
        if not self.load_patterns:
            raise ValueError(
                "No hay patrones de carga definidos. Agregue al menos uno para resolver el sistema.")

        # Inicializar análisis
        self.analysis = AnalysisManager(self)
        self.analysis.analyze_for_list_load_pattern(load_pattern_name)

        return self.results


    #! SETERS Y GETERS #######################################################
    def get_global_stiffness_matrix(
        self
    ) -> NDArray:
        """
        Obtiene la matriz de rigidez global del modelo.

        Returns:
            np.ndarray: Matriz de rigidez global.
        """
        return self.global_stiffness_matrix

    def get_global_load_vector(
        self,
        load_pattern_name: str
    ) -> NDArray:
        """
        Obtiene el vector de fuerzas global del modelo.

        Returns:
            np.ndarray: Vector de fuerzas global.
        """
        return self.global_load_vector.get(load_pattern_name)


    #! VISUALIZACION EN VENTANA INTERACTIVA ##################################
    def show(
        self
    ):
        """
        Muestra la interfaz gráfica para visualizar el modelo.
        """
        import time
        start_time = time.time()
        from milcapy.plotter.UIdisplay import main_window
        self.plotter = Plotter(self)
        self.plotter.initialize_plot()
        end_time = time.time()
        print(f"Tiempo de generacion de la visualizacion: {end_time - start_time}")
        main_window(self)

    #! RESULTADOS ###########################################################
    def get_results(
        self,
        load_pattern_name: str
    ):
        """
        Obtiene los resultados del análisis.

        Returns:
            Results: Objeto con los resultados del análisis.
        """
        return self.results.get(load_pattern_name)


    #! PLOT DE ELEMENTOS FINITOS #############################################

    def plot_model(
        self,
        load_pattern: Optional[str] = None,
        type: Optional[InternalForceType] = None,
    ):
        """
        Grafica el modelo estructural.
        """

        import matplotlib.pyplot as plt
        from milcapy.plotter.UIdisplay import MainWindow, QApplication, sys
        __Plotter = Plotter(self)#, load_pattern)
        __Plotter.initialize_plot()#type)

        # __Plotter.plotter_options.UI_axial = True
        # __Plotter.update_axial_force(visibility=True)

        plt.show()

        # app = QApplication.instance()
        # if app is None:  # Si no existe una instancia, se crea
        #     app = QApplication(sys.argv)
        # window = MainWindow(self)
        # window.show()



    def plot_field(
        self,
        field: Union[FieldTypeMembrane, str],
        cmap: str = "jet"
    ):
        """
        Grafica un campo de esfuerzos o deformaciones en el modelo.

        Args:
            field (Union[FieldType, str]): Campo de esfuerzos o deformaciones a graficar.
            cmap (str, opcional): Mapa de colores para la visualización. Default es "jet".

        Raises:
            ValueError: Si el campo no es válido.
        """
        # coordenadas nodales
        x = [node.vertex.x for node in self.nodes.values()]
        y = [node.vertex.y for node in self.nodes.values()]

        triangles = []
        node_values = {nid: [] for nid in self.nodes.keys()}

        for cst in self.csts.values():
            if isinstance(cst, MembraneTriangle):
                n1, n2, n3 = cst.node1, cst.node2, cst.node3
                triangles.append([n1.id - 1, n2.id - 1, n3.id - 1])

                # esfuerzos y deformaciones (al centroide)
                stresses = self.results[self.current_load_pattern].get_cst_stresses(
                    cst.id)
                strains = self.results[self.current_load_pattern].get_cst_strains(
                    cst.id)

                if field == FieldTypeMembrane.SX:
                    val = stresses[0]
                elif field == FieldTypeMembrane.SY:
                    val = stresses[1]
                elif field == FieldTypeMembrane.SXY:
                    val = stresses[2]
                elif field == FieldTypeMembrane.EX:
                    val = strains[0]
                elif field == FieldTypeMembrane.EY:
                    val = strains[1]
                elif field == FieldTypeMembrane.EXY:
                    val = strains[2]
                else:
                    val = None

                if val is not None:
                    # asignar valor del centroide a nodos
                    for n in [n1, n2, n3]:
                        node_values[n.id].append(val)

        # --- NUEVO: desplazamientos nodales ---
        if field in [FieldTypeMembrane.UX, FieldTypeMembrane.UY, FieldTypeMembrane.UMAG]:
            for nid, node in self.nodes.items():
                ux, uy = self.results[self.current_load_pattern].get_node_displacements(nid)[
                    :2]
                if field == FieldTypeMembrane.UX:
                    node_values[nid] = [ux]
                elif field == FieldTypeMembrane.UY:
                    node_values[nid] = [uy]
                elif field == FieldTypeMembrane.UMAG:
                    node_values[nid] = [np.sqrt(ux**2 + uy**2)]

        # promediar valores en los nodos
        nodal_field = np.zeros(len(self.nodes))
        for nid, vals in node_values.items():
            if vals:
                nodal_field[nid - 1] = np.mean(vals)

        triang = tri.Triangulation(x, y, triangles)

        # --- GRAFICADO CON tricontourf ---
        plt.figure(figsize=(15, 8))
        tcf = plt.tricontourf(triang, nodal_field,
                              levels=20, cmap=cmap)  # suavizado
        plt.colorbar(tcf, label=f"{field.value}")
        # --- GRAFICANDO LOS MIEMBROS ESTRUCTURALES ----
        for member in self.members.values():
            coord = np.hstack((member.node_i.vertex.coordinates,
                              member.node_j.vertex.coordinates))
            x = coord[[0, 2]]
            y = coord[[1, 3]]
            plt.plot(x, y, color="b")
        plt.axis('equal')
        plt.title(f"Mapa de calor - {field.value}")
        plt.xlabel("X")
        plt.ylabel("Y")
        plt.tight_layout()
        plt.show()

    def _plot_ux(self):
        self.plot_field(FieldTypeMembrane.UX)

    def _plot_uy(self):
        self.plot_field(FieldTypeMembrane.UY)

    def _plot_umag(self):
        self.plot_field(FieldTypeMembrane.UMAG)

    def _plot_sx(self):
        self.plot_field(FieldTypeMembrane.SX)

    def _plot_sy(self):
        self.plot_field(FieldTypeMembrane.SY)

    def _plot_sxy(self):
        self.plot_field(FieldTypeMembrane.SXY)

    def _plot_ex(self):
        self.plot_field(FieldTypeMembrane.EX)

    def _plot_ey(self):
        self.plot_field(FieldTypeMembrane.EY)

    def _plot_exy(self):
        self.plot_field(FieldTypeMembrane.EXY)


    #! METODOS PRIVADOS ######################################################
    def __id_verifier(self, ele_id: int, *id_nodes: int):
        """
        Verifica que el ID del elemento no exista y que los nodos existan.

        Args:
            ele_id (int): ID del elemento.
            *id_nodes (int): IDs de los nodos.

        Raises:
            ValueError: Si el ID del elemento ya existe o si no existen los nodos.
        """
        if (ele_id in self.members
            or ele_id in self.csts
            or ele_id in self.membrane_q3dof
                or ele_id in self.membrane_q2dof
                    or ele_id in self.trusses):
            raise ValueError(f"Ya existe un elemento con el ID {ele_id}")

        for node_id in id_nodes:
            if node_id not in self.nodes:
                raise ValueError(f"No existe un nodo con el ID {node_id}")

    def __get_vector(self, nA: Node, nB: Node) -> np.ndarray:
        """Vector de A hacia B."""
        return np.array([nB.x - nA.x, nB.y - nA.y], dtype=float)

    def __is_right_angle(self, v1: np.ndarray, v2: np.ndarray, tol: float = 1e-8) -> bool:
        """Verifica si dos vectores son ortogonales (producto punto ≈ 0)."""
        return abs(np.dot(v1, v2)) < tol

    def __is_same_length(self, v1: np.ndarray, v2: np.ndarray, tol: float = 1e-8) -> bool:
        """Verifica si dos vectores tienen la misma longitud."""
        return abs(np.linalg.norm(v1) - np.linalg.norm(v2)) < tol

    def __is_rectangular(self, n1_id: int, n2_id: int, n3_id: int, n4_id: int) -> bool:
        # """
        # Verifica si los nodos forman un rectángulo en sentido antihorario.
        # Orden esperado: n1 → n2 → n3 → n4.
        # """
        # try:
        #     n1, n2, n3, n4 = self.nodes[n1_id], self.nodes[n2_id], self.nodes[n3_id], self.nodes[n4_id]
        # except KeyError as e:
        #     raise ValueError(f"Nodo no encontrado: {e}")

        # # Vectores de lados
        # v12 = self.__get_vector(n1, n2)
        # v23 = self.__get_vector(n2, n3)
        # v34 = self.__get_vector(n3, n4)
        # v41 = self.__get_vector(n4, n1)

        # # Condiciones:
        # cond1 = self.__is_right_angle(v12, v23) and self.__is_right_angle(v23, v34) \
        #         and self.__is_right_angle(v34, v41) and self.__is_right_angle(v41, v12)

        # cond2 = self.__is_same_length(v12, v34) and self.__is_same_length(v23, v41)

        # return cond1 and cond2
        pass

    def __is_counterclockwise(self, *node_ids: int) -> bool:
        # """
        # Verifica si los nodos dados están en sentido antihorario (CCW).
        # Args:
        #     *node_ids: IDs de nodos en el orden del polígono.
        # Returns:
        #     bool: True si el polígono definido por los nodos está en sentido CCW.
        # """
        # if len(node_ids) < 3:
        #     raise ValueError("Se requieren al menos 3 nodos para definir un polígono.")

        # try:
        #     coords = [self.nodes[nid].vertex.coordinates for nid in node_ids]
        # except KeyError as e:
        #     raise ValueError(f"Nodo no encontrado: {e}")

        # area_signed = 0.0
        # n = len(coords)
        # for i in range(n):
        #     x1, y1 = coords[i]
        #     x2, y2 = coords[(i + 1) % n]  # siguiente nodo (cierra el polígono)
        #     area_signed += (x2 - x1) * (y2 + y1)  # variante del shoelace para orientación

        # # Otra opción equivalente:
        # # area_signed = sum(coords[i][0]*coords[(i+1)%n][1] - coords[(i+1)%n][0]*coords[i][1] for i in range(n))

        # return area_signed < 0  # <0 CCW, >0 CW (depende de convención)
        pass


from milcapy.postprocess.segment_member import BeamSeg
from milcapy.postprocess.CST_pp import PP_CST
from dataclasses import dataclass
from typing import TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    from milcapy.model.model import SystemMilcaModel
    from milcapy.core.results import Results


@dataclass
class PostProcessingOptions:
    """Opciones para el post-procesamiento de resultados estructurales.

    Attributes:
        factor (float): Factor de escala para la visualización de resultados.
        n (int): Número de puntos para discretizar los elementos.
    """
    factor: float
    n: int


class PostProcessing:   # para un solo load pattern
    """Clase para el post-procesamiento de resultados estructurales."""

    def __init__(
        self,
        model: "SystemMilcaModel",
        results: "Results",                 # ya viene con U, R del modelo para el LP activo
        options: "PostProcessingOptions",
        load_pattern_name: str
    ) -> None:
        """
        Inicializa el post-procesamiento para un sistema estructural.

        Args:
        model: Sistema estructural analizado
        results: Resultados del análisis
        options: Opciones de post-procesamiento
        load_pattern_name: Nombre del patrón de carga
    """
        self.model = model
        self.results = results
        self.options = options
        self.load_pattern_name = load_pattern_name

        self.reactions = self.results.model["reactions"]
        self.displacements = self.results.model["displacements"]

    def process_displacements_for_nodes(self) -> None:
        """Almacena las desplazamientos de los nodos en el objeto Results en sistema GLOBAL."""
        for id, node in self.model.nodes.items():
            if node.local_axis is None:
                array_displacements = self.displacements[node.dofs-1]
            if node.local_axis is not None:
                array_displacements = self.displacements[node.dofs-1]
                Tlg = node.local_axis.get_transformation_matrix()
                array_displacements = Tlg @ array_displacements
            self.results.set_node_displacements(id, array_displacements)

    def process_reactions_for_nodes(self) -> None:
        """Almacena las reacciones de los nodos en el objeto Results en sistema GLOBAL."""
        for id, node in self.model.nodes.items():
            if np.all(self.reactions[node.dofs - 1] == 0):
                pass
            else:
                if node.local_axis is None:
                    array_reactions = self.reactions[node.dofs-1]
                if node.local_axis is not None:
                    array_reactions = self.reactions[node.dofs-1]
                    Tlg = node.local_axis.get_transformation_matrix()
                    array_reactions = Tlg @ array_reactions
                self.results.set_node_reactions(id, array_reactions)

    def process_displacements_for_members(self) -> None:
        """Almacena las desplazamientos de los miembros en sistema LOCAL en el objeto Results."""
        for id, member in self.model.members.items():
            global_displacements = np.hstack((self.results.get_node_displacements(member.node_i.id), self.results.get_node_displacements(member.node_j.id)))
            array_displacements = np.dot(member.transformation_matrix(), global_displacements)
            self.results.set_member_displacements(id, array_displacements)

    def process_internal_forces_for_members(self) -> None:
        """Almacena las fuerzas internas de los miembros en sistema LOCAL en el objeto Results."""
        for id, member in self.model.members.items():
            local_displacements = self.results.get_member_displacements(id)
            load_vector = member.local_load_vector()
            stiffness_matrix = member.local_stiffness_matrix()
            array_internal_forces = np.dot(stiffness_matrix, local_displacements) - load_vector
            self.results.set_member_internal_forces(id, array_internal_forces)

    def process_displacements_for_cst(self) -> None:
        for id, cst in self.model.csts.items():
            global_displacements = np.hstack((self.results.get_node_displacements(cst.node1.id)[:2], self.results.get_node_displacements(cst.node2.id)[:2], self.results.get_node_displacements(cst.node3.id)[:2]))
            self.results.set_cst_displacements(id, global_displacements)

    def process_displacements_for_membrane_q3dof(self) -> None:
        for id, membrane_q3dof in self.model.membrane_q3dof.items():
            global_displacements = np.hstack((self.results.get_node_displacements(membrane_q3dof.node1.id), self.results.get_node_displacements(membrane_q3dof.node2.id), self.results.get_node_displacements(membrane_q3dof.node3.id), self.results.get_node_displacements(membrane_q3dof.node4.id)))
            self.results.set_membrane_q3dof_displacements(id, global_displacements)

    def process_displacements_for_membrane_q2dof(self) -> None:
        for id, membrane_q2dof in self.model.membrane_q2dof.items():
            global_displacements = np.hstack((self.results.get_node_displacements(membrane_q2dof.node1.id)[:2], self.results.get_node_displacements(membrane_q2dof.node2.id)[:2], self.results.get_node_displacements(membrane_q2dof.node3.id)[:2], self.results.get_node_displacements(membrane_q2dof.node4.id)[:2]))
            self.results.set_membrane_q2dof_displacements(id, global_displacements)

    def process_displacements_for_trusses(self) -> None:
        for id, truss in self.model.trusses.items():
            global_displacements = np.hstack((self.results.get_node_displacements(truss.node_i.id)[:2], self.results.get_node_displacements(truss.node_j.id)[:2]))
            Tlg = truss.transformation_matrix()
            local_displacements = Tlg @ global_displacements
            self.results.set_truss_displacements(id, local_displacements)

    def process_internal_forces_for_trusses(self) -> None:
        for id, truss in self.model.trusses.items():
            local_displacements = self.results.get_truss_displacements(id)
            load_vector = truss.q_local()
            stiffness_matrix = truss.local_stiffness_matrix()
            array_internal_forces = np.dot(stiffness_matrix, local_displacements) - load_vector
            self.results.set_truss_internal_forces(id, array_internal_forces)


    def post_process_for_members(self) -> None:
        """Almacena todos los resultados para cada miembro en el objeto Results (solo parte flexible)."""

        n = self.options.n
        calculator = BeamSeg()

        for id, member in self.model.members.items():

            result = self.results.get_results_member(id)
            calculator.process_builder(member, result, self.load_pattern_name)
            calculator.coefficients()

            if member.la or member.lb:
                # si hay end length offsets
                la = member.la or 0
                lb = member.lb or 0
                L = member.length()
                x_val = np.linspace(la, L-lb, n - 4)
                dx = L/100000
                x_val = np.hstack(([0, la-dx], x_val, [L-lb + dx, L]))
                self.results.set_x_val(id, x_val)
            else:
                x_val = np.linspace(0, member.length(), n)
                self.results.set_x_val(id, x_val)

            array_axial_force = np.zeros(n)
            array_axial_displacement = np.zeros(n)
            array_shear_force = np.zeros(n)
            array_bending_moment = np.zeros(n)
            array_slope = np.zeros(n)
            array_deflection = np.zeros(n)

            for i, x in enumerate(x_val):
                # Calcular fuerzas axiales
                array_axial_force[i] = calculator.axial(x)

                # Calcular los dezplazamientos axiales
                array_axial_displacement[i] = calculator.axial_displacement(x)

                # Calcular fuerzas cortantes
                array_shear_force[i] = calculator.shear(x)

                # Calcular momentos de flexión
                array_bending_moment[i] = calculator.moment(x)

                # Calcular pendientes
                array_slope[i] = calculator.slope(x)

                # Calcular deflexiones
                array_deflection[i] = calculator.deflection(x)

            self.results.set_member_axial_force(id, array_axial_force)
            self.results.set_member_axial_displacement(id, array_axial_displacement)
            self.results.set_member_shear_force(id, array_shear_force)
            self.results.set_member_bending_moment(id, array_bending_moment)
            self.results.set_member_slope(id, array_slope)
            self.results.set_member_deflection(id, array_deflection)


    def post_process_for_cst(self) -> None:
        """Almacacena los deformaciones unitarias y los esfuerzos en el centro del element"""
        calculator = PP_CST()

        for id, cst in self.model.csts.items():
            displacements = self.results.get_cst_displacements(id)
            calculator.process_builder(cst, displacements)

            strains = calculator.strains()
            stresses = calculator.stresses()
            self.results.set_cst_strains(id, strains)
            self.results.set_cst_stresses(id, stresses)

    def post_process_for_membrane_q3dof(self) -> None:
        pass

    def post_process_for_membrane_q2dof(self) -> None:
        pass

    def post_process_for_trusses(self) -> None:
        """Almacena todos los resultados para cada truss en el objeto Results."""

        n = self.options.n
        calculator = BeamSeg()

        for id, truss in self.model.trusses.items():
            results = self.results.get_results_truss(id)
            calculator.process_builder_for_truss(truss, results, self.load_pattern_name)
            calculator.coefficients()
            x_val = np.linspace(0, truss.length(), n)
            array_axial_force = np.zeros(n)
            array_axial_displacement = np.zeros(n)
            for i, x in enumerate(x_val):
                array_axial_force[i] = calculator.axial(x)
                array_axial_displacement[i] = calculator.axial_displacement(x)


            self.results.set_truss_axial_force(id, array_axial_force)
            self.results.set_truss_axial_displacement(id, array_axial_displacement)



from typing import Optional


class Material:
    """
    Clase base para representar un material estructural.

    Args:
        name (str): Nombre del material.
        modulus_elasticity (float): Módulo de elasticidad (Pa).
        poisson_ratio (float): Coeficiente de Poisson (-1 < v < 0.5).
        specific_weight (float): Peso específico o densidad (N/m³ o kg/m³ según el sistema de unidades).
    """

    def __init__(
        self,
        name: str,
        modulus_elasticity: float,
        poisson_ratio: float,
        specific_weight: float,
        alpha: Optional[float] = 1e-6,
        beta: Optional[float] = 1e-6,
        gamma: Optional[float] = 1e-6
        ) -> None:

        if modulus_elasticity <= 0:
            raise ValueError("El módulo de elasticidad (E) debe ser mayor que 0.")
        if not (0 <= poisson_ratio < 0.5):
            raise ValueError("El coeficiente de Poisson (v) debe estar en el rango (-1, 0.5).")
        if specific_weight < 0:
            raise ValueError("El peso específico o densidad (g) debe ser mayor o igual que 0.")

        self.name: str = name
        self.E: float = modulus_elasticity
        self.v: float = poisson_ratio
        self.g: float = specific_weight
        self.kg: float = 1 # Modificador de peso
        self.alpha: float = alpha # Coeficiente de expansión lineal
        self.beta: float = beta # Coeficiente de expansión superficial
        self.gamma: float = gamma # Coeficiente de expansión volumétrica

    @property
    def G(self) -> float:
        """
        Calcula el módulo de Corte (G) basado en el módulo de elasticidad y el coeficiente de Poisson.
        """
        return self.E / (2 * (1 + self.v))

    def __str__(self) -> str:
        return (f"{self.__class__.__name__}: {self.name}, "
                f"E={self.E:.2f}, v={self.v:.3f}, "
                f"G={self.G:.2f}, g={self.g:.2f}, "
                f"alpha={self.alpha:.6e}, beta={self.beta:.6e}, gamma={self.gamma:.6e}")


class ConcreteMaterial(Material):
    """Clase para representar materiales de concreto."""
    pass


class SteelMaterial(Material):
    """Clase para representar materiales de acero estructural."""
    pass


class GenericMaterial(Material):
    """Clase para representar materiales genéricos sin clasificación específica."""
    pass



"""
Módulo de cargas estructurales.

Proporciona clases y funciones para definir y manipular cargas en un modelo estructural, 
incluyendo cargas puntuales y patrones de carga.
"""

from ..loads.load import PointLoad, DistributedLoad, PrescribedDOF, ElasticSupport, LocalAxis, EndLengthOffset, CSTLoad
from ..loads.load_pattern import LoadPattern


from ast import Dict
from milcapy.core.node import Node
from milcapy.section.section import ShellSection
import numpy as np
from milcapy.loads.load import CSTLoad
from typing import Optional
from milcapy.utils.types import ConstitutiveModel

class MembraneTriangle:
    """
    Elemento de membrana triangular (elemento mas basico de FEM, constante).
    """
    def __init__(
        self,
        id: int,
        node1: Node,
        node2: Node,
        node3: Node,
        section: ShellSection,
        state: ConstitutiveModel,
        ) -> None:
        """
        Inicializa el elemento de membrana triangular.

        Args:
            id (int): Identificador del elemento.
            node1 (Node): Primer nodo.
            node2 (Node): Segundo nodo.
            node3 (Node): Tercer nodo.
            section (ShellSection): Sección del elemento tipo area (shell).
            state (ConstitutiveModel): Estado constitutivo del elemento.
        """
        self.id = id
        self.node1 = node1
        self.node2 = node2
        self.node3 = node3
        self.section = section
        self.state = state
        self.dofs = np.concatenate((node1.dofs[:2], node2.dofs[:2], node3.dofs[:2]))
        self.current_load_pattern: Optional[str] = None
        self.loads: Dict[str, CSTLoad] = {} # {lp_name: CSTLoad}

    @property
    def A(self) -> float:
        """
        Area del elemento.
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates

        a1 = x2*y3 - x3*y2
        a2 = x3*y1 - x1*y3
        a3 = x1*y2 - x2*y1

        return (a1 + a2 + a3)/2

    @property
    def B(self) -> np.ndarray:
        """
        Matriz de deformacion del elemento.
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates

        a1 = x2*y3 - x3*y2
        a2 = x3*y1 - x1*y3
        a3 = x1*y2 - x2*y1

        b1 = y2 - y3
        b2 = y3 - y1
        b3 = y1 - y2

        c1 = x3 - x2
        c2 = x1 - x3
        c3 = x2 - x1

        a = (a1 + a2 + a3)/2

        B = (1/(2*a))*np.array([
            [b1, 0,  b2, 0,  b3, 0],
            [0,  c1, 0,  c2, 0,  c3],
            [c1, b1, c2, b2, c3, b3],
        ])

        return B

    @property
    def D(self) -> np.ndarray:
        """
        Matriz constitutiva de esfuerzo plano.
        """
        v = self.section.v()
        E = self.section.E()
        # ! T E N S I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRESS:
            k = E/(1-v**2)
            return k*np.array([
                            [1, v, 0],
                            [v, 1, 0],
                            [0, 0, (1-v)/2],
                            ])

        # ! D E F O R M A C I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRAIN:
            k = (E*(1-v))/((1 + v)*(1 - 2*v))
            return k * np.array([
                [1,                 v/(1-v), 0],
                [v/(1-v),           1,       0],
                [0,                 0,       (1-2*v)/(2*(1-v))],
            ])

    def global_stiffness_matrix(self) -> np.ndarray:
        """
        Matriz de rigidez del elemento.
        """
        return self.B.T @ self.D.T @ self.B * self.section.t * self.A

    def set_load(self, load: CSTLoad) -> None:
        """
        Asigna una carga al elemento para el patrón de carga actual.
        """
        self.loads[self.current_load_pattern] = load

    def get_load(self) -> Optional[CSTLoad]:
        """
        Obtiene la carga asignada al elemento para el patron de carga actual.
        """
        return self.loads.get(self.current_load_pattern, None)

    def get_load_vector(self) -> Optional[np.ndarray]:
        """
        Obtiene el vector de cargas equivalentes en los nodos para el patrond e carga actual.
        """
        cstLoad = self.get_load()
        return cstLoad.Feq if cstLoad else np.zeros(6)


import numpy as np


def calculate_x_intersection(x1, y1, x2, y2):
    """Calcula la intersección con el eje X entre dos puntos"""
    if y1 == y2:
        return x1
    return x1 - y1 * (x2 - x1) / (y2 - y1)

def separate_areas(array, L, x_val=None):
    """Separa las áreas positivas y negativas con puntos de intersección"""
    if x_val is None:
        x_val = np.linspace(0, L, len(array))
    positive_areas = []
    negative_areas = []
    current_pos = []
    current_neg = []
    prev_sign = None

    for i, (x, y) in enumerate(zip(x_val, array)):
        current_sign = 'pos' if y >= 0 else 'neg'

        if i == 0:
            if current_sign == 'pos':
                current_pos.append([x, y])
            else:
                current_neg.append([x, y])
            prev_sign = current_sign
            continue

        if current_sign != prev_sign:
            # Calcular punto de intersección
            x_prev = x_val[i-1]
            y_prev = array[i-1]
            x_intersect = calculate_x_intersection(
                x_prev, y_prev, x, y)

            # Cerrar el área anterior
            if prev_sign == 'pos':
                current_pos.append([x_intersect, 0])
                positive_areas.append(current_pos)
                current_pos = []
            else:
                current_neg.append([x_intersect, 0])
                negative_areas.append(current_neg)
                current_neg = []

            # Iniciar nueva área
            if current_sign == 'pos':
                current_pos = [[x_intersect, 0], [x, y]]
            else:
                current_neg = [[x_intersect, 0], [x, y]]
            prev_sign = current_sign
        else:
            if current_sign == 'pos':
                current_pos.append([x, y])
            else:
                current_neg.append([x, y])
            prev_sign = current_sign

    # Añadir áreas restantes
    if current_pos:
        positive_areas.append(current_pos)
    if current_neg:
        negative_areas.append(current_neg)

    return positive_areas, negative_areas

def process_segments(segments, L):
    """Añade puntos en el eje X para segmentos en los bordes del dominio"""
    processed = []
    for seg in segments:
        new_seg = []
        # Verificar inicio
        if seg and seg[0][0] == 0 and seg[0][1] != 0:
            new_seg.append([0.0, 0.0])

        new_seg.extend(seg)

        # Verificar final
        if seg and seg[-1][0] == L and seg[-1][1] != 0:
            new_seg.append([L, 0.0])

        processed.append(new_seg)
    return processed



def redondear_si_decimal(x: float, ndigits: int = 2):
    """
    Redondea x a 'ndigits' decimales solo si x no es un entero exacto.
    
    Args:
        x (float | int): Número a evaluar.
        ndigits (int): Número de decimales a redondear si corresponde.
    
    Returns:
        float | int: x redondeado o mantenido.
    """
    if float(x).is_integer():
        return int(x)   # mantenerlo como entero
    else:
        return round(x, ndigits)



"""
==============================================
ELEMENTO CUADRILATERO MEMBRANA (Q4)
==============================================
|| Nodos                 || 4               ||
|| DOF por nodo          || 2               ||
|| Integracion           || Reducida 4P     ||
|| Shape Function        || bilineal        ||
|| Formulacion           || Isoparametrica  ||
|| clase de elemento     || Serendipitos    ||
|| Estado de programacion|| Listo           ||
==============================================
"""

from milcapy.core.node import Node
from milcapy.section.section import ShellSection
from milcapy.utils.types import ConstitutiveModel
import numpy as np

class MembraneQuad4:
    def __init__(
        self,
        id: int,
        node1: Node,
        node2: Node,
        node3: Node,
        node4: Node,
        section: ShellSection,
        state: ConstitutiveModel,
    ) -> None:
        self.id = id
        self.node1 = node1
        self.node2 = node2
        self.node3 = node3
        self.node4 = node4
        self.state = state
        self.section = section
        self.E = section.E()
        self.v = section.v()
        self.t = section.t
        self.dofs = np.concatenate((node1.dofs[:2], node2.dofs[:2], node3.dofs[:2], node4.dofs[:2]))

        r3 = np.sqrt(3)
        self.xi = [-1/r3, -1/r3, 1/r3, 1/r3]
        self.eta = [-1/r3, 1/r3, -1/r3, 1/r3]
        self.w = [1, 1, 1, 1]

    def shape_function(self, xi: float, eta: float) -> tuple[float, float, float, float]:
        """
        Calcula las funciones de forma en un punto (xi, eta).
        """
        N1 = 1/4 * (1-xi) * (1-eta)
        N2 = 1/4 * (1+xi) * (1-eta)
        N3 = 1/4 * (1+xi) * (1+eta)
        N4 = 1/4 * (1-xi) * (1+eta)
        return N1, N2, N3, N4

    def coordinates(self, xi: float, eta: float) -> tuple[float, float]:
        """
        Calcula las coordenadas en un punto (xi, eta).
        """
        N1,N2,N3,N4 = self.shape_function(xi, eta)
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates
        x = N1*x1 + N2*x2 + N3*x3 + N4*x4
        y = N1*y1 + N2*y2 + N3*y3 + N4*y4
        return x, y

    def Ia(self) -> np.ndarray:
        """
        Calcula la matriz unitaria ampliada.
        """
        Ia = np.array([
            [1, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 1, 1, 0]
        ])
        return Ia

    def J(self,xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz Jacobiana.
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates
        J11 = -x1*(1-eta) + x2*(1-eta) + x3*(1+eta) - x4*(1+eta)
        J12 = -y1*(1-eta) + y2*(1-eta) + y3*(1+eta) - y4*(1+eta)
        J21 = -x1*(1-xi) - x2*(1+xi) + x3*(1+xi) + x4*(1-xi)
        J22 = -y1*(1-xi) - y2*(1+xi) + y3*(1+xi) + y4*(1-xi)
        return np.array([[J11, J12], [J21, J22]]) * (1/4)

    def Gamma(self, xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz Jacobiana invertida.
        """
        J = self.J(xi, eta)
        gamma = np.linalg.inv(J)
        zero = np.zeros((2,2))
        Gamma = np.concatenate((
            np.concatenate((gamma, zero), axis=1),
            np.concatenate((zero, gamma), axis=1)
            ),
            axis=0)
        return Gamma

    def dN(self, xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz de derivadas de las funciones de forma.
        """
        dN = (1/4)*np.array([
            [eta-1, 0, 1-eta, 0, 1+eta, 0, -1-eta, 0],
            [xi-1,  0, -1-xi, 0, 1+xi,   0,  1-xi, 0],
            [0,     eta-1, 0, 1-eta, 0, 1+eta, 0, -1-eta],
            [0,     xi-1,  0, -1-xi, 0, 1+xi,   0,  1-xi]
        ])
        return dN


    def B(self,xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz de deformacion.
        """
        dN = self.dN(xi, eta)
        Gamma = self.Gamma(xi, eta)
        Ia = self.Ia()
        B = Ia @ Gamma @ dN
        return B

    def D(self) -> np.ndarray:
        """
        Matriz constitutiva de esfuerzo plano.
        """
        v = self.section.v()
        E = self.section.E()
        # ! T E N S I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRESS:
            k = E/(1-v**2)
            return k*np.array([
                            [1, v, 0],
                            [v, 1, 0],
                            [0, 0, (1-v)/2],
                            ])

        # ! D E F O R M A C I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRAIN:
            k = (E*(1-v))/((1 + v)*(1 - 2*v))
            return k * np.array([
                [1,                 v/(1-v), 0],
                [v/(1-v),           1,       0],
                [0,                 0,       (1-2*v)/(2*(1-v))],
            ])

    def phiKi(self, xi: float, eta: float):
        """
        Calcula la matriz funcional de rigidez.
        """
        detJ = np.linalg.det(self.J(xi, eta))
        B = self.B(xi, eta)
        D = self.D()
        t = self.t
        return B.T @ D @ B * detJ * t

    def global_stiffness_matrix(self):
        """
        Calcula la matriz rigidez global.
        """
        Ki = np.zeros((8,8))
        for i in range(len(self.xi)):
            Ki += self.phiKi(self.xi[i], self.eta[i]) * self.w[i]
        return Ki



if __name__ == "__main__":
    from milcapy.utils.geometry import Vertex
    E = 30*10**6 # psi
    v = 0.25
    t = 1 # in
    nd1 = Node(1, Vertex(3, 2))
    nd2 = Node(2, Vertex(5, 2))
    nd3 = Node(3, Vertex(5, 4))
    nd4 = Node(4, Vertex(3, 4))
    el = MembraneQuad4(1, nd1, nd2, nd3, nd4, E, v, t, ConstitutiveModel.PLANE_STRESS)
    k = 10**4
    print(el.global_stiffness_matrix()/k)











from typing import Dict, Tuple, List, TYPE_CHECKING
import numpy as np
from milcapy.postprocess.segment_member import deformed_shape
from milcapy.analysis.linear_static import LinearStaticAnalysis
from milcapy.utils import rotation_matrix

if TYPE_CHECKING:
    from milcapy.model.model import SystemMilcaModel

class PlotterValues:
    """
    Clase que prepara los valores calculados para ser representados gráficamente.
    Mantiene una referencia a datos estáticos compartidos entre todas las instancias.
    """

    _static_data: dict = None

    @classmethod
    def initialize_static_data(cls, system: 'SystemMilcaModel') -> None:
        """
        Inicializa los datos estáticos de la estructura una sola vez.

        los datos son de la forma:
        {
            'nodes': {node_id: (x, y)},
            'members': {member_id: ((x1, y1), (x2, y2))},
            'trusses': {member_id: ((x1, y1), (x2, y2))},
            'restraints': {node_id: (bool, bool, bool)}
        }
        """
        if cls._static_data is None:
            # Coordenadas de nodos
            nodes = {node.id: (node.vertex.x, node.vertex.y)
                    for node in system.nodes.values()}

            # Coordenadas de elementos
            members = {}
            for member in system.members.values():
                node_i = member.node_i
                node_j = member.node_j
                members[member.id] = (
                    (node_i.vertex.x, node_i.vertex.y),
                    (node_j.vertex.x, node_j.vertex.y)
                )

            # Coordenadas de trusses
            trusses = {}
            for truss in system.trusses.values():
                node_i = truss.node_i
                node_j = truss.node_j
                trusses[truss.id] = (
                    (node_i.vertex.x, node_i.vertex.y),
                    (node_j.vertex.x, node_j.vertex.y)
                )

            # Restricciones de nodos
            restraints = {}
            for node in system.nodes.values():
                if node.restraints != (False, False, False):
                    restraints[node.id] = node.restraints

            cls._static_data = {
                'nodes': nodes,
                'members': members,
                'trusses': trusses,
                'restraints': restraints
            }

    def __init__(
        self,
        model: 'SystemMilcaModel',
        current_load_pattern: str,
    ) -> None:
        """
        Inicializa los valores para un load pattern específico.

        Args:
            system: Sistema estructural analizado
            current_load_pattern: Nombre del load pattern
        """
        # Inicializar datos estáticos si aún no se ha hecho
        self.initialize_static_data(model)

        self.model = model
        self.current_load_pattern = current_load_pattern
        self.results = model.results[current_load_pattern]
        self.load_pattern = model.load_patterns.get(current_load_pattern, None)
        self.distributed_loads = {} # {member_id: {member_id: {qi, qj, pi, pj, mi, mj}}}
        self.point_loads       = {} # {node_id: {node_id: {fx, fy, mz}}}
        self.prescribed_dofs   = {} # {node_id: PrescribedDOF}
        self.elastic_supports  = {} # {node_id: ElasticSupport}
        self.cst               = {} # {cst_id: (coords, displacements)}
        self.membrane_q3dof       = {} # {membrane_q3dof_id: (coords, displacements)}
        self.membrane_q2dof      = {} # {membrane_q2dof_id: (coords, displacements)}
        self.trusses           = {} # {truss_id: (coords, displacements)}

        if not self.load_pattern:
            raise ValueError(
                f"Load pattern con nombre '{current_load_pattern}' no encontrado")

        # Procesamiento de datos dinámicos específicos del load pattern
        self._process_load_data()
        self._process_prescribed_dofs()
        self._process_elastic_supports()
        self._process_cst_element()
        self._process_membrane_q3dof()
        self._process_membrane_q2dof()
        self._process_trusses()

    def _process_load_data(self) -> None:
        """Procesa los datos de cargas para el load pattern actual."""
        # Cargas distribuidas
        for member_id, loads in self.load_pattern.distributed_loads.items():
            self.distributed_loads[member_id] = loads.to_dict()

        # Cargas puntuales
        for node_id, loads in self.load_pattern.point_loads.items():
            self.point_loads[node_id] = loads.to_dict()

    # Propiedades para acceder a datos estáticos
    @property
    def nodes(self) -> Dict[int, Tuple[float, float]]:
        """Devuelve las coordenadas de los nodos."""
        return self._static_data['nodes']

    @property
    def members(self) -> Dict[int, List[Tuple[float, float]]]:
        """Devuelve las coordenadas de los miembros."""
        return self._static_data['members']

    @property
    def restraints(self) -> Dict[int, Tuple[bool, bool, bool]]:
        """Devuelve las restricciones de los nodos."""
        return self._static_data['restraints']

    def rigid_deformed(self, member_id: int, escale: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:
        member = self.model.members[member_id]
        MT = member.transformation_matrix()
        arraydisp = self.results.members[member_id]['displacements']
        arraydisp = np.dot(MT.T, arraydisp)
        x_val = np.array([
            member.node_i.vertex.x + arraydisp[0] * escale,
            member.node_j.vertex.x + arraydisp[3] * escale
        ])
        y_val = np.array([
            member.node_i.vertex.y + arraydisp[1] * escale,
            member.node_j.vertex.y + arraydisp[4] * escale
        ])
        return x_val, y_val

    def get_deformed_shape(self, member_id: int, escale: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:

        """
        Obtiene la deformada de un miembro en coordenadas globales.
        """

        member = self.model.members[member_id]

        # Obtener la deformada en coordenadas LOCAL
        x_val, y_val = deformed_shape(member, self.results.members[member_id], escale)

        # Rotar el vector de deflexiones
        deformada_local = np.column_stack((x_val, y_val))
        deformada_global = np.dot(deformada_local, rotation_matrix(member.angle_x()).T) + member.node_i.vertex.coordinates

        x_val = deformada_global[:, 0]
        y_val = deformada_global[:, 1]

        return x_val, y_val

    def _process_prescribed_dofs(self) -> None:
        """Mapea los deslazamientos prescritos."""

        for node in self.model.nodes.values():
            PDOF = node.prescribed_dofs.get(self.model.current_load_pattern)
            if PDOF is not None:
                self.prescribed_dofs[node.id] = PDOF

    def _process_elastic_supports(self) -> None:
        """Mapea los apoyos elásticos."""
        for node in self.model.nodes.values():
            if node.elastic_supports is not None:
                self.elastic_supports[node.id] = node.elastic_supports

    def _process_cst_element(self) -> None:
        """
        Obtiene la coordenadas y desplazamientos de un CST en coordenadas globales.
        """
        for cst in self.model.csts.values():
            coordinates =  np.concatenate(([cst.node1.vertex.coordinates, cst.node2.vertex.coordinates, cst.node3.vertex.coordinates]))
            displacements = self.model.results[self.current_load_pattern].get_cst_displacements(cst.id)
            self.cst[cst.id] = (coordinates, displacements)

    def _process_membrane_q3dof(self) -> None:
        """
        Obtiene la coordenadas y desplazamientos de un MembraneQ6 en coordenadas globales.
        """
        for membrane_q3dof in self.model.membrane_q3dof.values():
            coordinates =  np.concatenate(([membrane_q3dof.node1.vertex.coordinates, membrane_q3dof.node2.vertex.coordinates, membrane_q3dof.node3.vertex.coordinates, membrane_q3dof.node4.vertex.coordinates]))
            displacements = self.model.results[self.current_load_pattern].get_membrane_q3dof_displacements(membrane_q3dof.id)
            self.membrane_q3dof[membrane_q3dof.id] = (coordinates, displacements)

    def _process_membrane_q2dof(self) -> None:
        """
        Obtiene la coordenadas y desplazamientos de un MembraneQ6i en coordenadas globales.
        """
        for membrane_q2dof in self.model.membrane_q2dof.values():
            coordinates =  np.concatenate(([membrane_q2dof.node1.vertex.coordinates, membrane_q2dof.node2.vertex.coordinates, membrane_q2dof.node3.vertex.coordinates, membrane_q2dof.node4.vertex.coordinates]))
            displacements = self.model.results[self.current_load_pattern].get_membrane_q2dof_displacements(membrane_q2dof.id)
            self.membrane_q2dof[membrane_q2dof.id] = (coordinates, displacements)

    def _process_trusses(self) -> None:
        """
        Obtiene la coordenadas y desplazamientos de un Truss en coordenadas globales.
        """
        for truss in self.model.trusses.values():
            coordinates =  np.concatenate((truss.node_i.vertex.coordinates, truss.node_j.vertex.coordinates))
            displacements = np.concatenate((self.results.get_node_displacements(truss.node_i.id)[:2], self.results.get_node_displacements(truss.node_j.id)[:2]))
            self.trusses[truss.id] = (coordinates, displacements)


import numpy as np


class Vertex:
    def __init__(
        self,
        x: float,
        y: float,
    ):
        self.coordinates: np.ndarray = np.array([x, y], dtype=np.float32)
        self.x: float = float(self.coordinates[0])
        self.y: float = float(self.coordinates[1])

    def distance_to(self, other: "Vertex") -> float:
        """Calcular la distancia entre dos vértices

        Args:
            other (Vertex): Otro vértice

        Returns:
            float: Distancia entre los vértices
        """
        return np.linalg.norm(self.coordinates - other.coordinates)

    def __str__(self) -> str:
        return f"Vertex({self.x}, {self.y})"


class MatrixException(Exception):
    """Excepción personalizada para errores en análisis matricial."""

    def __init__(self, type_: str, message: str):
        super().__init__(message)
        self.type = type_
        self.message = message


def rotation_matrix(angle: float) -> np.ndarray:
    """Crea una matriz de rotación 2x2.

    Args:
        angle (float): Ángulo en radianes.

    Returns:
        np.ndarray: Matriz de rotación 2x2.
    """
    s, c = np.sin(angle), np.cos(angle)
    return np.array([[c, -s], [s, c]])


def rotate_xy(a: np.ndarray, angle: float, x: float, y: float) -> np.ndarray:
    """Rota una matriz 2D alrededor del origen, en sentido antihorario.

    Args:
        a (np.ndarray): Matriz a rotar.
        angle (float): Ángulo en radianes.

    Returns:
        np.ndarray: Matriz rotada.
    """
    b = a - np.array([x, y])  # Centrar en el origen
    b = np.dot(b, rotation_matrix(angle).T)  # Aplicar rotación
    return b + np.array([x, y])  # Descentrar


def traslate_xy(a: np.ndarray, x: float, y: float) -> np.ndarray:
    """Traslada una matriz 2D.

    Args:
        a (np.ndarray): Matriz a trasladar.
        x (float): Traslación en x.
        y (float): Traslación en y.

    Returns:
        np.ndarray: Matriz trasladada.
    """
    return a + np.array([x, y])


def angle_x_axis(delta_x: float, delta_y: float) -> float:
    """Calcula el ángulo de un elemento respecto al eje x global usando NumPy.

    Args:
        delta_x (float): Longitud en dirección x.
        delta_y (float): Longitud en dirección y.

    Returns:
        float: Ángulo en radianes.
    """
    angle = np.arccos(delta_x / np.hypot(delta_x, delta_y))
    return 2 * np.pi - angle if delta_y < 0 else angle

def vertex_range(v1: Vertex, v2: Vertex, n: int) -> list:
    """
    Genera una lista de n+1 vértices que se encuentran en el rango entre v1 y v2.

    Args:
        v1 (Vertex): Vértice inicial.
        v2 (Vertex): Vértice final.
        n (int): Número de vértices a generar.

    Returns:
        list: Lista de vértices.
    """
    dv = v2 - v1
    return [v1 + dv * i / n for i in range(n + 1)]


from matplotlib.lines import Line2D
from milcapy.utils import rotate_xy
from math import sin, cos, pi
import numpy as np

lw=0.7

def draw_roller_support(ax, x, y, size=0.1, theta=0, color='#20dd75', lw=lw):
    """
    Dibuja un apoyo movil sin fijacion de rotacion.
    """
    a = size
    theta = np.deg2rad(theta)
    vertex = np.array([
        (x, y), #1
        (x, y - a / 2),#2
        (x - a / 2, y - a/2),#3
        (x + a / 2, y - a/2),#4
        (x, y - a / 2),#2
        (np.nan, np.nan),#5
        (x, y - a),#5
        (x - a/2, y - a),#5
        (x + a/2, y - a),#6
    ])
    rotated = rotate_xy(vertex, theta, x, y)
    line = Line2D(rotated[:, 0], rotated[:, 1], color=color, linewidth=lw)
    ax.add_line(line)
    return line

def draw_pinned_support(ax, x, y, size=0.1, theta=0, color='#20dd75', lw=lw):
    """
    Dibuja un apoyo fijo.
    """
    a = size
    theta = np.deg2rad(theta)
    vertex = np.array([
        (x, y-a/4),
        (x - a / 2, y - a),
        (x + a / 2, y - a),
        (x, y-a/4),
        (x, y - a/4),
        (x, y)
    ])
    rotated = rotate_xy(vertex, theta, x, y)
    line = Line2D(rotated[:, 0], rotated[:, 1], color=color, linewidth=lw)
    ax.add_line(line)
    return line

def draw_fixed_support(ax, x, y, size=0.1, theta=0, color='#20dd75', lw=lw):
    """
    Dibuja un apoyo empotrado.
    """
    a = size
    theta = np.deg2rad(theta)
    vertex = np.array([
        (x, y - a/3),
        (x - a/2, y - a/3),
        (x - a/2, y - a),
        (x + a/2, y - a),
        (x + a/2, y - a/3),
        (x, y - a/3),
        (x, y)
    ])
    rotated = rotate_xy(vertex, theta, x, y)
    line = Line2D(rotated[:, 0], rotated[:, 1], color=color, linewidth=lw)
    ax.add_line(line)
    return line

def draw_roller_support2(ax, x, y, size=0.1, theta=0, color='#20dd75', lw=lw):
    """
    Dibuja un apoyo movil con fijacion de rotacion.
    """
    a = size
    theta = np.deg2rad(theta)
    vertex = np.array([
        (x, y),
        (x - a/2, y - 3*a/4),
        (x + a/2, y - 3*a/4),
        (x, y),
        (np.nan, np.nan),
        (x-a/2, y - a),
        (x+a/2, y - a)
    ])
    rotated = rotate_xy(vertex, theta, x, y)
    line = Line2D(rotated[:, 0], rotated[:, 1], color=color, linewidth=lw)
    ax.add_line(line)
    return line





def support_ttt(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo empotrado.
    """
    a = size
    vertex = np.array([
        (x, y - a/3),
        (x - a/2, y - a/3),
        (x - a/2, y - a),
        (x + a/2, y - a),
        (x + a/2, y - a/3),
        (x, y - a/3),
        (x, y)
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_ttf(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo fijo sobre el eje X.
    """
    a = size

    vertex = np.array([
        (x, y-a/4),
        (x - a / 2, y - a),
        (x + a / 2, y - a),
        (x, y-a/4),
        (x, y - a/4),
        (x, y)
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_tft(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo movil sobre el eje Y.
    """
    a = size
    vertex = np.array([
        (x, y),
        (x - a / 2, y),
        (x - a / 2, y - a/2),
        (x - a / 2, y + a/2),
        (x - a / 2, y),
        (np.nan, np.nan),
        (x - a, y),
        (x - a, y - a/2),
        (x - a, y + a/2)
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_ftt(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo movil sobre el eje X.
    """
    a = size
    vertex = np.array([
        (x, y), #1
        (x, y - a / 2),#2
        (x - a / 2, y - a/2),#3
        (x + a / 2, y - a/2),#4
        (x, y - a / 2),#2
        (np.nan, np.nan),#5
        (x, y - a),#5
        (x - a/2, y - a),#5
        (x + a/2, y - a),#6
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_tff(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo fijo sobre el eje X y movil sobre el eje Y libre a rotacion.
    """
    a = size
    vertex = np.array([
        (x, y),#1
        (x - 3*a / 4, y - a/2),#2
        (x - 3*a / 4, y + a/2),#3
        (x, y),#1
        (np.nan, np.nan),#4
        (x-a, y - a/2),#5
        (x-a, y + a/2)#6
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_ftf(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo movil sobre el eje X y fijo sobre el eje Y libre a rotacion.
    """
    a = size
    vertex = np.array([
        (x, y),
        (x - a/2, y - 3*a/4),
        (x + a/2, y - 3*a/4),
        (x, y),
        (np.nan, np.nan),
        (x-a/2, y - a),
        (x+a/2, y - a)
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_fff(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Punto sin restricciones, puede representarse vacío.
    """
    return

def support_fft(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo movil en el eje X e Y y fijo a rotacion.
    """
    a = size
    vertex = np.array([
        (x-a/2, y),
        (x+a/2, y),
        (x, y),
        (x, y-a/2),
        (x, y+a/2),
        (x, y),
        (x+cos(pi/4)*a/2, y + sin(pi/4)*a/2),
        (x+cos(5*pi/4)*a/2, y+sin(5*pi/4)*a/2),
        (x, y),
        (x + cos(3*pi/4)*a/2, y+sin(3*pi/4)*a/2),
        (x+cos(7*pi/4)*a/2, y +sin(7*pi/4)*a/2)
    ])
    vertex = rotate_xy(vertex, theta, x, y)
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_kt(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo elastico (resorte a desplazamiento).
    """
    a = size
    k1 = 1/4
    k2 = 3/16
    vertex = [
        (-a, -k1*a),
        (-a, k1*a),
        (-a, 0),
        (-3*a/4, 0),
        (-3*a/4, k2*a),
        (-a/2, -k2*a),
        (-a/2, k2*a),
        (-a/2, k2*a),
        (-a/4, -k2*a),
        (-a/4, 0),
        (0, 0)
    ]
    vertex = np.array(vertex)

    # Rotación
    rad = np.deg2rad(theta)
    rot_matrix = np.array([[np.cos(rad), -np.sin(rad)],
                           [np.sin(rad),  np.cos(rad)]])
    vertex = vertex @ rot_matrix.T

    # Traslación
    vertex[:, 0] += x
    vertex[:, 1] += y

    # Dibujar
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line

def support_kr(ax, x, y, size=0.1, color='#20dd75', lw=lw, zorder=3, theta=0):
    """
    Dibuja un apoyo elastico (resorte a rotacion).
    """
    a = size
    k1 = 1/4
    n_vueltas = 2
    ri = 0.03 * a
    rf = a/2 + ri - 0.25 * a
    puntos_por_vuelta = 20

    # Espiral
    t = np.linspace(0, 2 * np.pi * n_vueltas, n_vueltas * puntos_por_vuelta)
    r = np.linspace(ri, rf, len(t))
    xs = r * np.cos(t) - a/2 - ri
    ys = r * np.sin(t)

    # Vértices iniciales
    vertex = [
        (-a, -k1*a),
        (-a,  k1*a),
        (-a,  0),
        (-a/2, 0)
    ]

    # Añadir espiral
    vertex.extend(list(zip(xs, ys)))
    vertex.append((0, 0))

    # Convertir a numpy
    vertex = np.array(vertex)

    # Rotar vertices
    rad = np.deg2rad(theta)
    rot_matrix = np.array([[np.cos(rad), -np.sin(rad)],
                           [np.sin(rad),  np.cos(rad)]])
    vertex = vertex @ rot_matrix.T

    # Trasladar al punto (x, y)
    vertex[:, 0] += x
    vertex[:, 1] += y

    # Dibujar
    line = Line2D(vertex[:, 0], vertex[:, 1], color=color, linewidth=lw, zorder=zorder)
    ax.add_line(line)
    return line
